@@ -1,9 +1,9 @@
 #
 # Parser for XML-RPC call and response
-# 
+#
 # Copyright (C) 2001, 2002, 2003 by Michael Neumann (mneumann@ntecs.de)
 #
-# $Id: parser.rb 2062 2006-06-10 19:14:15Z headius $
+# $Id: parser.rb 22784 2009-03-06 03:56:38Z nobu $
 #
 
 
@@ -60,7 +60,7 @@ module XMLRPC
       @faultCode   = faultCode
       @faultString = faultString
     end
-    
+
     # returns a hash
     def to_h
       {"faultCode" => @faultCode, "faultString" => @faultString}
@@ -77,7 +77,7 @@ module XMLRPC
       when "0" then false
       when "1" then true
       else
-        raise "RPC-value of type boolean is wrong" 
+        raise "RPC-value of type boolean is wrong"
       end
     end
 
@@ -92,7 +92,7 @@ module XMLRPC
         if $7
           ofs = $8.to_i*3600 + $9.to_i*60
           ofs = -ofs if $7=='+'
-          utc = Time.utc(a.reverse) + ofs
+          utc = Time.utc(*a) + ofs
           a = [ utc.year, utc.month, utc.day, utc.hour, utc.min, utc.sec ]
         end
         XMLRPC::DateTime.new(*a)
@@ -106,7 +106,7 @@ module XMLRPC
         if $7
           ofs = $8.to_i*3600 + $9.to_i*60
           ofs = -ofs if $7=='+'
-          utc = Time.utc(a.reverse) + ofs
+          utc = Time.utc(*a) + ofs
           a = [ utc.year, utc.month, utc.day, utc.hour, utc.min, utc.sec ]
         end
         XMLRPC::DateTime.new(*a)
@@ -122,7 +122,7 @@ module XMLRPC
     def self.struct(hash)
       # convert to marhalled object
       klass = hash["___class___"]
-      if klass.nil? or Config::ENABLE_MARSHALLING == false 
+      if klass.nil? or Config::ENABLE_MARSHALLING == false
         hash
       else
         begin
@@ -130,9 +130,9 @@ module XMLRPC
           klass.split("::").each {|const| mod = mod.const_get(const.strip)}
 
           obj = mod.allocate
-          
+
           hash.delete "___class___"
-          hash.each {|key, value| 
+          hash.each {|key, value|
             obj.instance_variable_set("@#{ key }", value) if key =~ /^([\w_][\w_0-9]*)$/
           }
           obj
@@ -143,11 +143,11 @@ module XMLRPC
     end
 
     def self.fault(hash)
-      if hash.kind_of? Hash and hash.size == 2 and 
-        hash.has_key? "faultCode" and hash.has_key? "faultString" and 
+      if hash.kind_of? Hash and hash.size == 2 and
+        hash.has_key? "faultCode" and hash.has_key? "faultString" and
         hash["faultCode"].kind_of? Integer and hash["faultString"].kind_of? String
 
-        XMLRPC::FaultException.new(hash["faultCode"], hash["faultString"]) 
+        XMLRPC::FaultException.new(hash["faultCode"], hash["faultString"])
       else
         raise "wrong fault-structure: #{hash.inspect}"
       end
@@ -160,11 +160,11 @@ module XMLRPC
     class AbstractTreeParser
 
       def parseMethodResponse(str)
-	methodResponse_document(createCleanedTree(str))
+        methodResponse_document(createCleanedTree(str))
       end
 
       def parseMethodCall(str)
-	methodCall_document(createCleanedTree(str))
+        methodCall_document(createCleanedTree(str))
       end
 
       private
@@ -174,205 +174,205 @@ module XMLRPC
       # and all comments
       #
       def removeWhitespacesAndComments(node)
-	remove = []
-	childs = node.childNodes.to_a
-	childs.each do |nd|
-	  case _nodeType(nd)
-	  when :TEXT
+        remove = []
+        childs = node.childNodes.to_a
+        childs.each do |nd|
+          case _nodeType(nd)
+          when :TEXT
             # TODO: add nil?
             unless %w(i4 int boolean string double dateTime.iso8601 base64).include? node.nodeName
 
-               if node.nodeName == "value" 
+               if node.nodeName == "value"
                  if not node.childNodes.to_a.detect {|n| _nodeType(n) == :ELEMENT}.nil?
-                   remove << nd if nd.nodeValue.strip == "" 
+                   remove << nd if nd.nodeValue.strip == ""
                  end
                else
                  remove << nd if nd.nodeValue.strip == ""
                end
-	    end
-	  when :COMMENT
-	    remove << nd
-	  else
-	    removeWhitespacesAndComments(nd)
-	  end 
-	end
+            end
+          when :COMMENT
+            remove << nd
+          else
+            removeWhitespacesAndComments(nd)
+          end
+        end
 
-	remove.each { |i| node.removeChild(i) }
+        remove.each { |i| node.removeChild(i) }
       end
 
 
       def nodeMustBe(node, name)
-	cmp = case name
-	when Array 
-	  name.include?(node.nodeName)
-	when String
-	  name == node.nodeName
-	else
-	  raise "error"
-	end  
+        cmp = case name
+        when Array
+          name.include?(node.nodeName)
+        when String
+          name == node.nodeName
+        else
+          raise "error"
+        end
 
-	if not cmp then
-	  raise "wrong xml-rpc (name)"
-	end
+        if not cmp then
+          raise "wrong xml-rpc (name)"
+        end
 
-	node
+        node
       end
 
       #
       # returns, when successfully the only child-node
       #
       def hasOnlyOneChild(node, name=nil)
-	if node.childNodes.to_a.size != 1
-	  raise "wrong xml-rpc (size)"
-	end
-	if name != nil then
-	  nodeMustBe(node.firstChild, name)
-	end
+        if node.childNodes.to_a.size != 1
+          raise "wrong xml-rpc (size)"
+        end
+        if name != nil then
+          nodeMustBe(node.firstChild, name)
+        end
       end
 
 
       def assert(b)
-	if not b then
-	  raise "assert-fail" 
-	end
+        if not b then
+          raise "assert-fail"
+        end
       end
 
       # the node `node` has empty string or string
       def text_zero_one(node)
-	nodes = node.childNodes.to_a.size
+        nodes = node.childNodes.to_a.size
 
-	if nodes == 1
-	  text(node.firstChild)
-	elsif nodes == 0
-	  ""
-	else
-	  raise "wrong xml-rpc (size)"
-	end
+        if nodes == 1
+          text(node.firstChild)
+        elsif nodes == 0
+          ""
+        else
+          raise "wrong xml-rpc (size)"
+        end
       end
-     
+
 
       def integer(node)
-	#TODO: check string for float because to_i returnsa
-	#      0 when wrong string
-	 nodeMustBe(node, %w(i4 int))    
-	hasOnlyOneChild(node)
-	
-	Convert.int(text(node.firstChild))
+        #TODO: check string for float because to_i returnsa
+        #      0 when wrong string
+         nodeMustBe(node, %w(i4 int))
+        hasOnlyOneChild(node)
+
+        Convert.int(text(node.firstChild))
       end
 
       def boolean(node)
-	nodeMustBe(node, "boolean")    
-	hasOnlyOneChild(node)
-	
+        nodeMustBe(node, "boolean")
+        hasOnlyOneChild(node)
+
         Convert.boolean(text(node.firstChild))
       end
 
       def v_nil(node)
         nodeMustBe(node, "nil")
-	assert( node.childNodes.to_a.size == 0 )
+        assert( node.childNodes.to_a.size == 0 )
         nil
       end
 
       def string(node)
-	nodeMustBe(node, "string")    
-	text_zero_one(node)
+        nodeMustBe(node, "string")
+        text_zero_one(node)
       end
 
       def double(node)
-	#TODO: check string for float because to_f returnsa
-	#      0.0 when wrong string
-	nodeMustBe(node, "double")    
-	hasOnlyOneChild(node)
-	
-	Convert.double(text(node.firstChild))
+        #TODO: check string for float because to_f returnsa
+        #      0.0 when wrong string
+        nodeMustBe(node, "double")
+        hasOnlyOneChild(node)
+
+        Convert.double(text(node.firstChild))
       end
 
       def dateTime(node)
-	nodeMustBe(node, "dateTime.iso8601")
-	hasOnlyOneChild(node)
-	
+        nodeMustBe(node, "dateTime.iso8601")
+        hasOnlyOneChild(node)
+
         Convert.dateTime( text(node.firstChild) )
       end
 
       def base64(node)
-	nodeMustBe(node, "base64")
-	#hasOnlyOneChild(node)
-	 
+        nodeMustBe(node, "base64")
+        #hasOnlyOneChild(node)
+
         Convert.base64(text_zero_one(node))
       end
 
       def member(node)
-	nodeMustBe(node, "member")
-	assert( node.childNodes.to_a.size == 2 ) 
+        nodeMustBe(node, "member")
+        assert( node.childNodes.to_a.size == 2 )
 
-	[ name(node[0]), value(node[1]) ]
+        [ name(node[0]), value(node[1]) ]
       end
 
       def name(node)
-	nodeMustBe(node, "name")
-	#hasOnlyOneChild(node)
-	text_zero_one(node) 
+        nodeMustBe(node, "name")
+        #hasOnlyOneChild(node)
+        text_zero_one(node)
       end
 
       def array(node)
-	nodeMustBe(node, "array")
-	hasOnlyOneChild(node, "data") 
-	data(node.firstChild)  
+        nodeMustBe(node, "array")
+        hasOnlyOneChild(node, "data")
+        data(node.firstChild)
       end
 
       def data(node)
-	nodeMustBe(node, "data")
+        nodeMustBe(node, "data")
 
-	node.childNodes.to_a.collect do |val|
-	  value(val)
-	end 
+        node.childNodes.to_a.collect do |val|
+          value(val)
+        end
       end
 
       def param(node)
-	nodeMustBe(node, "param")
-	hasOnlyOneChild(node, "value")
-	value(node.firstChild) 
+        nodeMustBe(node, "param")
+        hasOnlyOneChild(node, "value")
+        value(node.firstChild)
       end
- 
-      def methodResponse(node)
-	nodeMustBe(node, "methodResponse")
-	hasOnlyOneChild(node, %w(params fault))
-	child = node.firstChild
 
-	case child.nodeName
-	when "params"
-	  [ true, params(child,false) ] 
-	when "fault"
-	  [ false, fault(child) ]
-	else
-	  raise "unexpected error"
-	end
+      def methodResponse(node)
+        nodeMustBe(node, "methodResponse")
+        hasOnlyOneChild(node, %w(params fault))
+        child = node.firstChild
+
+        case child.nodeName
+        when "params"
+          [ true, params(child,false) ]
+        when "fault"
+          [ false, fault(child) ]
+        else
+          raise "unexpected error"
+        end
 
       end
 
       def methodName(node)
-	nodeMustBe(node, "methodName")
-	hasOnlyOneChild(node)
-	text(node.firstChild) 
+        nodeMustBe(node, "methodName")
+        hasOnlyOneChild(node)
+        text(node.firstChild)
       end
 
       def params(node, call=true)
-	nodeMustBe(node, "params")
+        nodeMustBe(node, "params")
 
-	if call 
-	  node.childNodes.to_a.collect do |n|
-	    param(n)
-	  end
-	else # response (only one param)
-	  hasOnlyOneChild(node)
-	  param(node.firstChild)
-	end
+        if call
+          node.childNodes.to_a.collect do |n|
+            param(n)
+          end
+        else # response (only one param)
+          hasOnlyOneChild(node)
+          param(node.firstChild)
+        end
       end
 
       def fault(node)
-	nodeMustBe(node, "fault")
-	hasOnlyOneChild(node, "value")
-	f = value(node.firstChild) 
+        nodeMustBe(node, "fault")
+        hasOnlyOneChild(node, "value")
+        f = value(node.firstChild)
         Convert.fault(f)
       end
 
@@ -380,76 +380,76 @@ module XMLRPC
 
       # _nodeType is defined in the subclass
       def text(node)
-	assert( _nodeType(node) == :TEXT )
-	assert( node.hasChildNodes == false )
-	assert( node.nodeValue != nil )
+        assert( _nodeType(node) == :TEXT )
+        assert( node.hasChildNodes == false )
+        assert( node.nodeValue != nil )
 
-	node.nodeValue.to_s
+        node.nodeValue.to_s
       end
 
       def struct(node)
-	nodeMustBe(node, "struct")    
+        nodeMustBe(node, "struct")
 
-	hash = {}
-	node.childNodes.to_a.each do |me|
-	  n, v = member(me)  
-	  hash[n] = v
-	end 
+        hash = {}
+        node.childNodes.to_a.each do |me|
+          n, v = member(me)
+          hash[n] = v
+        end
 
         Convert.struct(hash)
-     end
+      end
 
 
       def value(node)
-	nodeMustBe(node, "value")
-	nodes = node.childNodes.to_a.size
-        if nodes == 0 
+        nodeMustBe(node, "value")
+        nodes = node.childNodes.to_a.size
+        if nodes == 0
           return ""
-        elsif nodes > 1 
-	  raise "wrong xml-rpc (size)"
+        elsif nodes > 1
+          raise "wrong xml-rpc (size)"
         end
 
-	child = node.firstChild
+        child = node.firstChild
 
-	case _nodeType(child)
-	when :TEXT
+        case _nodeType(child)
+        when :TEXT
           text_zero_one(node)
-	when :ELEMENT
-	  case child.nodeName
-	  when "i4", "int"        then integer(child)
-	  when "boolean"          then boolean(child)
-	  when "string"           then string(child)
-	  when "double"           then double(child)
-	  when "dateTime.iso8601" then dateTime(child)
-	  when "base64"           then base64(child)
-	  when "struct"           then struct(child)
-	  when "array"            then array(child) 
-          when "nil"              
+        when :ELEMENT
+          case child.nodeName
+          when "i4", "int"        then integer(child)
+          when "boolean"          then boolean(child)
+          when "string"           then string(child)
+          when "double"           then double(child)
+          when "dateTime.iso8601" then dateTime(child)
+          when "base64"           then base64(child)
+          when "struct"           then struct(child)
+          when "array"            then array(child)
+          when "nil"
             if Config::ENABLE_NIL_PARSER
               v_nil(child)
             else
               raise "wrong/unknown XML-RPC type 'nil'"
             end
-	  else 
-	    raise "wrong/unknown XML-RPC type"
-	  end
-	else
-	  raise "wrong type of node"
-	end
+          else
+            raise "wrong/unknown XML-RPC type"
+          end
+        else
+          raise "wrong type of node"
+        end
 
       end
 
       def methodCall(node)
-	nodeMustBe(node, "methodCall")
-	assert( (1..2).include?( node.childNodes.to_a.size ) ) 
-	name = methodName(node[0])
+        nodeMustBe(node, "methodCall")
+        assert( (1..2).include?( node.childNodes.to_a.size ) )
+        name = methodName(node[0])
 
-	if node.childNodes.to_a.size == 2 then
-	  pa = params(node[1])
-	else # no parameters given
-	  pa = []
-	end
-	[name, pa]
+        if node.childNodes.to_a.size == 2 then
+          pa = params(node[1])
+        else # no parameters given
+          pa = []
+        end
+        [name, pa]
       end
 
     end # module TreeParserMixin
@@ -461,7 +461,7 @@ module XMLRPC
         raise "No valid method response!" if parser.method_name != nil
         if parser.fault != nil
           # is a fault structure
-          [false, parser.fault] 
+          [false, parser.fault]
         else
           # is a normal return value
           raise "Missing return value!" if parser.params.size == 0
@@ -508,7 +508,7 @@ module XMLRPC
           @value = nil
         when "nil"
           raise "wrong/unknown XML-RPC type 'nil'" unless Config::ENABLE_NIL_PARSER
-          @value = :nil 
+          @value = :nil
         when "array"
           @val_stack << @values
           @values = []
@@ -517,7 +517,7 @@ module XMLRPC
           @name = []
 
           @structs << @struct
-          @struct = {} 
+          @struct = {}
         end
       end
 
@@ -538,7 +538,7 @@ module XMLRPC
           @value = Convert.base64(@data)
         when "value"
           @value = @data if @value.nil?
-          @values << (@value == :nil ? nil : @value) 
+          @values << (@value == :nil ? nil : @value)
         when "array"
           @value = @values
           @values = @val_stack.pop
@@ -548,9 +548,9 @@ module XMLRPC
           @name = @names.pop
           @struct = @structs.pop
         when "name"
-          @name[0] = @data 
+          @name[0] = @data
         when "member"
-          @struct[@name[0]] = @values.pop 
+          @struct[@name[0]] = @values.pop
 
         when "param"
           @params << @values[0]
@@ -560,7 +560,7 @@ module XMLRPC
           @fault = Convert.fault(@values[0])
 
         when "methodName"
-          @method_name = @data 
+          @method_name = @data
         end
 
         @data = nil
@@ -592,7 +592,7 @@ module XMLRPC
         @parser_class = XMLRPCParser
       end
 
-      class XMLRPCParser 
+      class XMLRPCParser
         include StreamParserMixin
 
         def parse(str)
@@ -620,9 +620,9 @@ module XMLRPC
       def initialize
         require "xmltreebuilder"
 
-        # The new XMLParser library (0.6.2+) uses a slightly different DOM implementation. 
+        # The new XMLParser library (0.6.2+) uses a slightly different DOM implementation.
         # The following code removes the differences between both versions.
-        if defined? XML::DOM::Builder 
+        if defined? XML::DOM::Builder
           return if defined? XML::DOM::Node::DOCUMENT # code below has been already executed
           klass = XML::DOM::Node
           klass.const_set("DOCUMENT", klass::DOCUMENT_NODE)
@@ -635,34 +635,34 @@ module XMLRPC
       private
 
       def _nodeType(node)
-	tp = node.nodeType
-	if tp == XML::SimpleTree::Node::TEXT then :TEXT
-	elsif tp == XML::SimpleTree::Node::COMMENT then :COMMENT 
-	elsif tp == XML::SimpleTree::Node::ELEMENT then :ELEMENT 
-	else :ELSE
-	end
+        tp = node.nodeType
+        if tp == XML::SimpleTree::Node::TEXT then :TEXT
+        elsif tp == XML::SimpleTree::Node::COMMENT then :COMMENT
+        elsif tp == XML::SimpleTree::Node::ELEMENT then :ELEMENT
+        else :ELSE
+        end
       end
 
 
       def methodResponse_document(node)
-	assert( node.nodeType == XML::SimpleTree::Node::DOCUMENT )
-	hasOnlyOneChild(node, "methodResponse")
-	
-	methodResponse(node.firstChild)
+        assert( node.nodeType == XML::SimpleTree::Node::DOCUMENT )
+        hasOnlyOneChild(node, "methodResponse")
+
+        methodResponse(node.firstChild)
       end
 
       def methodCall_document(node)
-	assert( node.nodeType == XML::SimpleTree::Node::DOCUMENT )
-	hasOnlyOneChild(node, "methodCall")
-	
-	methodCall(node.firstChild)
+        assert( node.nodeType == XML::SimpleTree::Node::DOCUMENT )
+        hasOnlyOneChild(node, "methodCall")
+
+        methodCall(node.firstChild)
       end
 
       def createCleanedTree(str)
-	doc = XML::SimpleTreeBuilder.new.parse(str)
-	doc.documentElement.normalize
-	removeWhitespacesAndComments(doc)
-	doc
+        doc = XML::SimpleTreeBuilder.new.parse(str)
+        doc.documentElement.normalize
+        removeWhitespacesAndComments(doc)
+        doc
       end
 
     end # class XMLParser
@@ -676,21 +676,21 @@ module XMLRPC
       private
 
       def _nodeType(node)
-	node.nodeType
+        node.nodeType
       end
 
       def methodResponse_document(node)
-	methodResponse(node)
+        methodResponse(node)
       end
 
       def methodCall_document(node)
-	methodCall(node)
+        methodCall(node)
       end
 
       def createCleanedTree(str)
-        doc = ::NQXML::TreeParser.new(str).document.rootNode 
-	removeWhitespacesAndComments(doc)
-	doc
+        doc = ::NQXML::TreeParser.new(str).document.rootNode
+        removeWhitespacesAndComments(doc)
+        doc
       end
 
     end # class NQXMLTreeParser
@@ -701,7 +701,7 @@ module XMLRPC
         @parser_class = StreamListener
       end
 
-      class StreamListener 
+      class StreamListener
         include StreamParserMixin
 
         alias :tag_start :startElement
@@ -715,8 +715,8 @@ module XMLRPC
 
         def parse(str)
           parser = REXML::Document.parse_stream(str, self)
-       end
-      end 
+        end
+      end
 
     end
     # ---------------------------------------------------------------------------
@@ -751,7 +751,7 @@ module XMLRPC
           startElement(name)
           endElement(name)
         end
-       
+
         def on_chardata(str)
           character(str)
         end
@@ -784,7 +784,7 @@ module XMLRPC
         # valid_name?
         # valid_chardata?
         # valid_char?
-        # parse_error  
+        # parse_error
 
       end
     end
@@ -792,8 +792,8 @@ module XMLRPC
     XMLParser   = XMLTreeParser
     NQXMLParser = NQXMLTreeParser
 
-    Classes = [XMLStreamParser, XMLTreeParser, 
-               NQXMLStreamParser, NQXMLTreeParser, 
+    Classes = [XMLStreamParser, XMLTreeParser,
+               NQXMLStreamParser, NQXMLTreeParser,
                REXMLStreamParser, XMLScanStreamParser]
 
     # yields an instance of each installed parser
