@@ -48,12 +48,15 @@ import java.io.Reader;
 import java.nio.channels.Channels;
 import java.nio.channels.FileChannel;
 import java.nio.channels.FileLock;
+import java.util.zip.ZipEntry;
+import java.util.zip.ZipFile;
 
 import org.jruby.anno.JRubyClass;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.anno.JRubyModule;
 import org.jruby.ext.posix.util.Platform;
 import org.jruby.runtime.Block;
+import org.jruby.runtime.ClassIndex;
 import org.jruby.runtime.ObjectAllocator;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.Visibility;
@@ -61,6 +64,7 @@ import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.encoding.EncodingCapable;
 import org.jruby.util.ByteList;
 import org.jruby.util.io.DirectoryAsFileException;
+import org.jruby.util.io.PermissionDeniedException;
 import org.jruby.util.io.Stream;
 import org.jruby.util.io.ChannelStream;
 import org.jruby.util.io.ModeFlags;
@@ -166,7 +170,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         super(runtime, runtime.getFile());
         this.path = path;
         try {
-            this.openFile.setMainStream(new ChannelStream(runtime, new ChannelDescriptor(Channels.newChannel(in), getNewFileno(), new FileDescriptor())));
+            this.openFile.setMainStream(ChannelStream.open(runtime, new ChannelDescriptor(Channels.newChannel(in), getNewFileno(), new FileDescriptor())));
         } catch (InvalidValueException ex) {
             throw runtime.newErrnoEINVALError();
         }
@@ -196,6 +200,10 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         // Create Constants class
         RubyModule constants = fileClass.defineModuleUnder("Constants");
         runtime.setFile(fileClass);
+
+        fileClass.index = ClassIndex.FILE;
+        fileClass.setReifiedClass(RubyFile.class);
+
         RubyString separator = runtime.newString("/");
         ThreadContext context = runtime.getCurrentContext();
         
@@ -549,12 +557,21 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             // TODO: check if too many open files, GC and try again
 
             return descriptor;
+        } catch (PermissionDeniedException pde) {
+            // PDException can be thrown only when creating the file and
+            // permission is denied.  See JavaDoc of PermissionDeniedException.
+            throw getRuntime().newErrnoEACCESError(path);
         } catch (FileNotFoundException fnfe) {
-            throw getRuntime().newErrnoENOENTError();
+            // FNFException can be thrown in both cases, when the file
+            // is not found, or when permission is denied.
+            if (Ruby.isSecurityRestricted() || new File(path).exists()) {
+                throw getRuntime().newErrnoEACCESError(path);
+            }
+            throw getRuntime().newErrnoENOENTError(path);
         } catch (DirectoryAsFileException dafe) {
             throw getRuntime().newErrnoEISDirError();
         } catch (FileExistsException fee) {
-            throw getRuntime().newErrnoEEXISTError("file exists: " + path);
+            throw getRuntime().newErrnoEEXISTError(path);
         } catch (IOException ioe) {
             throw getRuntime().newIOErrorFromException(ioe);
         }
@@ -589,15 +606,17 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             return stream;
         } catch (BadDescriptorException e) {
             throw getRuntime().newErrnoEBADFError();
+        } catch (PermissionDeniedException pde) {
+            // PDException can be thrown only when creating the file and
+            // permission is denied.  See JavaDoc of PermissionDeniedException.
+            throw getRuntime().newErrnoEACCESError(path);
         } catch (FileNotFoundException ex) {
             // FNFException can be thrown in both cases, when the file
             // is not found, or when permission is denied.
             if (Ruby.isSecurityRestricted() || new File(path).exists()) {
-                throw getRuntime().newErrnoEACCESError(
-                        "Permission denied - " + path);
+                throw getRuntime().newErrnoEACCESError(path);
             }
-            throw getRuntime().newErrnoENOENTError(
-                    "File not found - " + path);
+            throw getRuntime().newErrnoENOENTError(path);
         } catch (DirectoryAsFileException ex) {
             throw getRuntime().newErrnoEISDirError();
         } catch (FileExistsException ex) {
@@ -609,7 +628,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         } catch (PipeException ex) {
             throw getRuntime().newErrnoEPIPEError();
         } catch (SecurityException ex) {
-            throw getRuntime().newErrnoEACCESError("Permission denied - " + path);
+            throw getRuntime().newErrnoEACCESError(path);
         }
     }
 
@@ -618,7 +637,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         int mode = (int) arg.convertToInteger().getLongValue();
 
         if (!new File(path).exists()) {
-            throw context.getRuntime().newErrnoENOENTError("No such file or directory - " + path);
+            throw context.getRuntime().newErrnoENOENTError(path);
         }
 
         return context.getRuntime().newFixnum(context.getRuntime().getPosix().chmod(path, mode));
@@ -637,7 +656,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         }
 
         if (!new File(path).exists()) {
-            throw context.getRuntime().newErrnoENOENTError("No such file or directory - " + path);
+            throw context.getRuntime().newErrnoENOENTError(path);
         }
 
         return context.getRuntime().newFixnum(context.getRuntime().getPosix().chown(path, owner, group));
@@ -658,7 +677,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         int mode = (int) arg.convertToInteger().getLongValue();
 
         if (!new File(path).exists()) {
-            throw context.getRuntime().newErrnoENOENTError("No such file or directory - " + path);
+            throw context.getRuntime().newErrnoENOENTError(path);
         }
 
         return context.getRuntime().newFixnum(context.getRuntime().getPosix().lchmod(path, mode));
@@ -678,7 +697,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         }
 
         if (!new File(path).exists()) {
-            throw context.getRuntime().newErrnoENOENTError("No such file or directory - " + path);
+            throw context.getRuntime().newErrnoENOENTError(path);
         }
 
         return context.getRuntime().newFixnum(context.getRuntime().getPosix().lchown(path, owner, group));
@@ -735,7 +754,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
     public IRubyObject truncate(ThreadContext context, IRubyObject arg) {
         RubyInteger newLength = arg.convertToInteger();
         if (newLength.getLongValue() < 0) {
-            throw context.getRuntime().newErrnoEINVALError("invalid argument: " + path);
+            throw context.getRuntime().newErrnoEINVALError(path);
         }
         try {
             openFile.checkWritable(context.getRuntime());
@@ -862,7 +881,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             RubyString filename = get_path(context, args[i]);
             
             if (!RubyFileTest.exist_p(filename, filename).isTrue()) {
-                throw runtime.newErrnoENOENTError("No such file or directory - " + filename);
+                throw runtime.newErrnoENOENTError(filename.toString());
             }
             
             boolean result = 0 == runtime.getPosix().chmod(filename.getUnicodeValue(), (int)mode.getLongValue());
@@ -892,7 +911,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             RubyString filename = get_path(context, args[i]);
 
             if (!RubyFileTest.exist_p(filename, filename).isTrue()) {
-                throw runtime.newErrnoENOENTError("No such file or directory - " + filename);
+                throw runtime.newErrnoENOENTError(filename.toString());
             }
             
             boolean result = 0 == runtime.getPosix().chown(filename.getUnicodeValue(), owner, group);
@@ -1289,8 +1308,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         ByteList pattern = args[0].convertToString().getByteList();
         ByteList path = get_path(context, args[1]).getByteList();
 
-        if (org.jruby.util.Dir.fnmatch(pattern.bytes, pattern.begin, pattern.begin+pattern.realSize, 
-                                       path.bytes, path.begin, path.begin+path.realSize, flags) == 0) {
+        if (org.jruby.util.Dir.fnmatch(pattern.getUnsafeBytes(), pattern.getBegin(), pattern.getBegin()+pattern.getRealSize(), path.getUnsafeBytes(), path.getBegin(), path.getBegin()+path.getRealSize(), flags) == 0) {
             return context.getRuntime().getTrue();
         }
         return context.getRuntime().getFalse();
@@ -1372,36 +1390,24 @@ public class RubyFile extends RubyIO implements EncodingCapable {
     @JRubyMethod(name = "lstat", required = 1, meta = true)
     public static IRubyObject lstat(ThreadContext context, IRubyObject recv, IRubyObject filename) {
         String f = get_path(context, filename).getUnicodeValue();
-        if(f.startsWith("file:") && f.indexOf('!') != -1) {
-            f = f.substring(5, f.indexOf("!"));
-        }
         return context.getRuntime().newFileStat(f, true);
     }
 
     @JRubyMethod(name = "stat", required = 1, meta = true)
     public static IRubyObject stat(ThreadContext context, IRubyObject recv, IRubyObject filename) {
         String f = get_path(context, filename).getUnicodeValue();
-        if(f.startsWith("file:") && f.indexOf('!') != -1) {
-            f = f.substring(5, f.indexOf("!"));
-        }
         return context.getRuntime().newFileStat(f, false);
     }
 
     @JRubyMethod(name = "atime", required = 1, meta = true)
     public static IRubyObject atime(ThreadContext context, IRubyObject recv, IRubyObject filename) {
         String f = get_path(context, filename).getUnicodeValue();
-        if(f.startsWith("file:") && f.indexOf('!') != -1) {
-            f = f.substring(5, f.indexOf("!"));
-        }
         return context.getRuntime().newFileStat(f, false).atime();
     }
 
     @JRubyMethod(name = "ctime", required = 1, meta = true)
     public static IRubyObject ctime(ThreadContext context, IRubyObject recv, IRubyObject filename) {
         String f = get_path(context, filename).getUnicodeValue();
-        if(f.startsWith("file:") && f.indexOf('!') != -1) {
-            f = f.substring(5, f.indexOf("!"));
-        }
         return context.getRuntime().newFileStat(f, false).ctime();
     }
 
@@ -1415,7 +1421,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             RubyString filename = get_path(context, args[i]);
             
             if (!RubyFileTest.exist_p(filename, filename).isTrue()) {
-                throw runtime.newErrnoENOENTError("No such file or directory - " + filename);
+                throw runtime.newErrnoENOENTError(filename.toString());
             }
             
             boolean result = 0 == runtime.getPosix().lchmod(filename.getUnicodeValue(), (int)mode.getLongValue());
@@ -1447,24 +1453,21 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         return runtime.newFixnum(count);
     }
 
-    @JRubyMethod(required = 2, meta = true)
+    @JRubyMethod(required = 2, meta = true, backtrace = true)
     public static IRubyObject link(ThreadContext context, IRubyObject recv, IRubyObject from, IRubyObject to) {
         Ruby runtime = context.getRuntime();
         RubyString fromStr = RubyString.stringValue(from);
         RubyString toStr = RubyString.stringValue(to);
-        
-        try {
-            if (runtime.getPosix().link(
-                    fromStr.getUnicodeValue(),toStr.getUnicodeValue()) == -1) {
-                // FIXME: When we get JNA3 we need to properly write this to errno.
-                throw runtime.newErrnoEEXISTError("File exists - " 
-                               + fromStr + " or " + toStr);
-            }
-        } catch (java.lang.UnsatisfiedLinkError ule) {
-            throw runtime.newNotImplementedError("link() function is unimplemented on this machine");
+
+        int ret = runtime.getPosix().link(fromStr.getUnicodeValue(), toStr.getUnicodeValue());
+        if (ret != 0) {
+            // In most cases, when there is an error during the call,
+            // the POSIX handler throws an exception, but not in case
+            // with pure Java POSIX layer (when native support is disabled),
+            // so we deal with it like this:
+            throw runtime.newErrnoEEXISTError(fromStr + " or " + toStr);
         }
-        
-        return runtime.newFixnum(0);
+        return runtime.newFixnum(ret);
     }
 
     @JRubyMethod(name = "mtime", required = 1, meta = true)
@@ -1486,8 +1489,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         JRubyFile newFile = JRubyFile.create(runtime.getCurrentDirectory(), newNameJavaString);
         
         if (!oldFile.exists() || !newFile.getParentFile().exists()) {
-            throw runtime.newErrnoENOENTError("No such file or directory - " + oldNameJavaString +
-                    " or " + newNameJavaString);
+            throw runtime.newErrnoENOENTError(oldNameJavaString + " or " + newNameJavaString);
         }
 
         JRubyFile dest = JRubyFile.create(runtime.getCurrentDirectory(), newNameJavaString);
@@ -1507,8 +1509,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             return RubyFixnum.zero(runtime);
         }
 
-        throw runtime.newErrnoEACCESError("Permission denied - " + oldNameJavaString + " or " +
-                newNameJavaString);
+        throw runtime.newErrnoEACCESError(oldNameJavaString + " or " + newNameJavaString);
     }
     
     @JRubyMethod(required = 1, meta = true)
@@ -1530,8 +1531,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             if (runtime.getPosix().symlink(
                     fromStr.getUnicodeValue(), tovalue) == -1) {
                 // FIXME: When we get JNA3 we need to properly write this to errno.
-                throw runtime.newErrnoEEXISTError("File exists - " 
-                               + fromStr + " or " + toStr);
+                throw runtime.newErrnoEEXISTError(fromStr + " or " + toStr);
             }
         } catch (java.lang.UnsatisfiedLinkError ule) {
             throw runtime.newNotImplementedError("symlink() function is unimplemented on this machine");
@@ -1548,11 +1548,11 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             String realPath = runtime.getPosix().readlink(path.convertToString().getUnicodeValue());
         
             if (!RubyFileTest.exist_p(recv, path).isTrue()) {
-                throw runtime.newErrnoENOENTError("No such file or directory - " + path);
+                throw runtime.newErrnoENOENTError(path.toString());
             }
         
             if (!RubyFileTest.symlink_p(recv, path).isTrue()) {
-                throw runtime.newErrnoEINVALError("invalid argument - " + path);
+                throw runtime.newErrnoEINVALError(path.toString());
             }
         
             if (realPath == null) {
@@ -1582,12 +1582,11 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         }
 
         if (!testFile.exists()) {
-            throw runtime.newErrnoENOENTError(
-                    "No such file or directory - " + filename.getByteList().toString());
+            throw runtime.newErrnoENOENTError(filename.getByteList().toString());
         }
 
         if (newLength.getLongValue() < 0) {
-            throw runtime.newErrnoEINVALError("invalid argument: " + filename);
+            throw runtime.newErrnoEINVALError(filename.toString());
         }
 
         IRubyObject[] args = new IRubyObject[] { filename, runtime.newString("r+") };
@@ -1632,7 +1631,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             JRubyFile fileToTouch = JRubyFile.create(runtime.getCurrentDirectory(),filename.getUnicodeValue());
             
             if (!fileToTouch.exists()) {
-                throw runtime.newErrnoENOENTError(" No such file or directory - \"" + filename + "\"");
+                throw runtime.newErrnoENOENTError(filename.toString());
             }
 
             runtime.getPosix().utimes(fileToTouch.getAbsolutePath(), atimeval, mtimeval);
@@ -1654,17 +1653,39 @@ public class RubyFile extends RubyIO implements EncodingCapable {
             // Broken symlinks considered by exists() as non-existing,
             // so we need to check for symlinks explicitly.
             if (!lToDelete.exists() && !isSymlink) {
-                throw runtime.newErrnoENOENTError(" No such file or directory - \"" + filename + "\"");
+                throw runtime.newErrnoENOENTError(filename.toString());
             }
 
             if (!lToDelete.delete()) {
-                throw runtime.newErrnoEACCESError("Permission denied - \"" + filename + "\"");
+                throw runtime.newErrnoEACCESError(filename.toString());
             }
         }
         
         return runtime.newFixnum(args.length);
     }
 
+    public static ZipEntry getFileEntry(ZipFile zf, String path) throws IOException {
+        ZipEntry entry = zf.getEntry(path);
+        if (entry == null) {
+            // try canonicalizing the path to eliminate . and .. (JRUBY-4760)
+            entry = zf.getEntry(new File("/" + path).getCanonicalPath().substring(1));
+        }
+        return entry;
+    }
+
+    public static ZipEntry getDirOrFileEntry(ZipFile zf, String path) throws IOException {
+        ZipEntry entry = zf.getEntry(path + "/"); // first try as directory
+        if (entry == null) {
+            // try canonicalizing the path to eliminate . and .. (JRUBY-4760)
+            entry = zf.getEntry(new File("/" + path + "/").getCanonicalPath().substring(1));
+            if (entry == null) {
+                // try as file
+                entry = getFileEntry(zf, path);
+            }
+        }
+        return entry;
+    }
+
     /**
      * Extract a timeval (an array of 2 longs: seconds and microseconds from epoch) from
      * an IRubyObject.
@@ -1698,7 +1719,7 @@ public class RubyFile extends RubyIO implements EncodingCapable {
         JRubyFile file = JRubyFile.create(runtime.getCurrentDirectory(), path);
         
         if (!file.exists()) {
-            throw runtime.newErrnoENOENTError("No such file or directory - " + path);
+            throw runtime.newErrnoENOENTError(path);
         }
         
         return runtime.newTime(file.lastModified());
