@@ -33,6 +33,7 @@ package org.jruby;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.anno.JRubyModule;
 import org.jruby.common.IRubyWarnings.ID;
+import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 
@@ -40,12 +41,18 @@ import org.jruby.runtime.builtin.IRubyObject;
  * GC (Garbage Collection) Module
  *
  * Note: Since we rely on Java's memory model we can't provide the
- * kind of control over garbage collection that MRI provides.
+ * kind of control over garbage collection that MRI provides.  Also note
+ * that since all Ruby libraries make GC assumptions based on MRI's GC
+ * that we decided to no-op explicit collection through these APIs.
+ * You can use Java Integration in your libraries to force a Java
+ * GC (assuming you really want to).
  *
- * @author Anders
  */
 @JRubyModule(name="GC")
 public class RubyGC {
+    private static volatile boolean gcDisabled = false;
+    private static volatile boolean stress = false;
+
     public static RubyModule createGCModule(Ruby runtime) {
         RubyModule result = runtime.defineModule("GC");
         runtime.setGC(result);
@@ -56,26 +63,48 @@ public class RubyGC {
     }
 
     @JRubyMethod(module = true, visibility = Visibility.PRIVATE)
-    public static IRubyObject start(IRubyObject recv) {
-        System.gc();
-        return recv.getRuntime().getNil();
+    public static IRubyObject start(ThreadContext context, IRubyObject recv) {
+        return context.getRuntime().getNil();
     }
 
     @JRubyMethod
-    public static IRubyObject garbage_collect(IRubyObject recv) {
-        System.gc();
-        return recv.getRuntime().getNil();
+    public static IRubyObject garbage_collect(ThreadContext context, IRubyObject recv) {
+        return context.getRuntime().getNil();
     }
 
     @JRubyMethod(module = true, visibility = Visibility.PRIVATE)
-    public static IRubyObject enable(IRubyObject recv) {
-        recv.getRuntime().getWarnings().warn(ID.EMPTY_IMPLEMENTATION, "GC.enable will not work on JRuby", "GC.enable");
-        return recv.getRuntime().getNil();
+    public static IRubyObject enable(ThreadContext context, IRubyObject recv) {
+        Ruby runtime = context.getRuntime();
+        emptyImplementationWarning(runtime, "GC.enable");
+        boolean old = gcDisabled;
+        gcDisabled = false;
+        return runtime.newBoolean(old);
     }
 
     @JRubyMethod(module = true, visibility = Visibility.PRIVATE)
-    public static IRubyObject disable(IRubyObject recv) {
-        recv.getRuntime().getWarnings().warn(ID.EMPTY_IMPLEMENTATION, "GC.disable will not work on JRuby", "GC.disable");
-        return recv.getRuntime().getNil();
+    public static IRubyObject disable(ThreadContext context, IRubyObject recv) {
+        Ruby runtime = context.getRuntime();
+        emptyImplementationWarning(runtime, "GC.disable");
+        boolean old = gcDisabled;
+        gcDisabled = true;
+        return runtime.newBoolean(old);
+    }
+
+    @JRubyMethod(module = true, visibility = Visibility.PRIVATE)
+    public static IRubyObject stress(ThreadContext context, IRubyObject recv) {
+        return context.getRuntime().newBoolean(stress);
+    }
+
+    @JRubyMethod(name = "stress=", module = true, visibility = Visibility.PRIVATE)
+    public static IRubyObject stress_set(ThreadContext context, IRubyObject recv, IRubyObject arg) {
+        Ruby runtime = context.getRuntime();
+        emptyImplementationWarning(runtime, "GC.stress=");
+        stress = arg.isTrue();
+        return runtime.newBoolean(stress);
+    }
+
+    private static void emptyImplementationWarning(Ruby runtime, String name) {
+        runtime.getWarnings().warn(ID.EMPTY_IMPLEMENTATION,
+                name + " does nothing on JRuby", name);
     }
 }
