@@ -199,7 +199,7 @@ public class RubyGlobal {
         runtime.defineVariable(new StringGlobalVariable(runtime, "$\\", runtime.getNil()));
         runtime.defineVariable(new StringGlobalVariable(runtime, "$,", runtime.getNil()));
 
-        runtime.defineVariable(new LineNumberGlobalVariable(runtime, "$.", RubyFixnum.one(runtime)));
+        runtime.defineVariable(new LineNumberGlobalVariable(runtime, "$."));
         runtime.defineVariable(new LastlineGlobalVariable(runtime, "$_"));
         runtime.defineVariable(new LastExitStatusVariable(runtime, "$?"));
 
@@ -264,7 +264,7 @@ public class RubyGlobal {
         // On platforms without a c-library accessable through JNA, getpid will return hashCode 
         // as $$ used to. Using $$ to kill processes could take down many runtimes, but by basing
         // $$ on getpid() where available, we have the same semantics as MRI.
-        runtime.getGlobalVariables().defineReadonly("$$", new ValueAccessor(runtime.newFixnum(runtime.getPosix().getpid())));
+        runtime.getGlobalVariables().defineReadonly("$$", new PidAccessor(runtime));
 
         // after defn of $stderr as the call may produce warnings
         defineGlobalEnvConstants(runtime);
@@ -409,14 +409,21 @@ public class RubyGlobal {
     // Accessor methods.
 
     private static class LineNumberGlobalVariable extends GlobalVariable {
-        public LineNumberGlobalVariable(Ruby runtime, String name, RubyFixnum value) {
-            super(runtime, name, value);
+        public LineNumberGlobalVariable(Ruby runtime, String name) {
+            super(runtime, name, null);
         }
 
         @Override
         public IRubyObject set(IRubyObject value) {
-            RubyArgsFile.setCurrentLineNumber(runtime.getGlobalVariables().get("$<"), value);
-            return super.set(value);
+            int line = (int)value.convertToInteger().getLongValue();
+            runtime.setCurrentLine(line);
+            RubyArgsFile.setCurrentLineNumber(runtime.getArgsFile(), line);
+            return value;
+        }
+
+        @Override
+        public IRubyObject get() {
+            return runtime.newFixnum(runtime.getCurrentLine());
         }
     }
 
@@ -657,4 +664,25 @@ public class RubyGlobal {
             return runtime.getLoadService().getLoadedFeatures();
         }
     }
+
+    /**
+     * A special accessor for getpid, to avoid loading the posix subsystem until
+     * it is needed.
+     */
+    private static final class PidAccessor implements IAccessor {
+        private final Ruby runtime;
+        private IRubyObject pid = null;
+
+        public PidAccessor(Ruby runtime) {
+            this.runtime = runtime;
+        }
+
+        public IRubyObject getValue() {
+            return pid != null ? pid : (pid = runtime.newFixnum(runtime.getPosix().getpid()));
+        }
+
+        public IRubyObject setValue(IRubyObject newValue) {
+            throw runtime.newRuntimeError("cannot assign to $$");
+        }
+    }
 }
