@@ -91,7 +91,7 @@ public class RubyMarshal {
                 throw recv.getRuntime().newTypeError("Instance of IO needed");
             }
             if (args.length == 3) {
-                depthLimit = (int) ((RubyFixnum) args[2]).getLongValue();
+                depthLimit = (int) args[2].convertToInteger().getLongValue();
             }
         }
 
@@ -100,11 +100,15 @@ public class RubyMarshal {
                 dumpToStream(objectToDump, outputStream(io), depthLimit);
                 return io;
             }
-			ByteArrayOutputStream stringOutput = new ByteArrayOutputStream();
-			dumpToStream(objectToDump, stringOutput, depthLimit);
-
-            return RubyString.newString(recv.getRuntime(), new ByteList(stringOutput.toByteArray(),false));
+            
+            ByteArrayOutputStream stringOutput = new ByteArrayOutputStream();
+            boolean taint = dumpToStream(objectToDump, stringOutput, depthLimit);
 
+            RubyString result = RubyString.newString(recv.getRuntime(), new ByteList(stringOutput.toByteArray(),false));
+            if(taint) {
+                result.setTaint(true);
+            }
+            return result;
         } catch (IOException ioe) {
             throw recv.getRuntime().newIOErrorFromException(ioe);
         }
@@ -113,9 +117,6 @@ public class RubyMarshal {
 
     private static OutputStream outputStream(IRubyObject out) {
         setBinmodeIfPossible(out);
-        if (out instanceof RubyIO) {
-            return ((RubyIO) out).getOutStream();
-        }
         return new IOOutputStream(out);
     }
 
@@ -127,6 +128,8 @@ public class RubyMarshal {
 
     @JRubyMethod(name = {"load", "restore"}, required = 1, optional = 1, frame = true, module = true)
     public static IRubyObject load(ThreadContext context, IRubyObject recv, IRubyObject[] args, Block unusedBlock) {
+        IRubyObject in = null;
+        IRubyObject proc = null;
         try {
             if (args.length < 1) {
                 throw recv.getRuntime().newArgumentError("wrong number of arguments (0 for 1)");
@@ -136,9 +139,6 @@ public class RubyMarshal {
             	throw recv.getRuntime().newArgumentError("wrong number of arguments (" + args.length + " for 2)");
             }
             
-            IRubyObject in = null;
-            IRubyObject proc = null;
-
             switch (args.length) {
             case 2:
             	proc = args[1];
@@ -147,21 +147,27 @@ public class RubyMarshal {
             }
 
             InputStream rawInput;
+            boolean tainted = false;
             if (in != null && in.respondsTo("read")) {
+                tainted = in.isTaint();
                 rawInput = inputStream(in);
             } else if (in != null && in.respondsTo("to_str")) {
+                tainted = in.isTaint();
                 RubyString inString = (RubyString) RuntimeHelpers.invoke(context, in, "to_str");
                 ByteList bytes = inString.getByteList();
-                rawInput = new ByteArrayInputStream(bytes.unsafeBytes(), bytes.begin(), bytes.length());
+                rawInput = new ByteArrayInputStream(bytes.getUnsafeBytes(), bytes.begin(), bytes.length());
             } else {
                 throw recv.getRuntime().newTypeError("instance of IO needed");
             }
             
-            UnmarshalStream input = new UnmarshalStream(recv.getRuntime(), rawInput, proc);
+            UnmarshalStream input = new UnmarshalStream(recv.getRuntime(), rawInput, proc, tainted);
 
             return input.unmarshalObject();
 
         } catch (EOFException ee) {
+          if (in != null && in.respondsTo("to_str"))
+            throw recv.getRuntime().newArgumentError("marshal data too short");
+          else
             throw recv.getRuntime().newEOFError();
         } catch (IOException ioe) {
             throw recv.getRuntime().newIOErrorFromException(ioe);
@@ -170,16 +176,14 @@ public class RubyMarshal {
 
     private static InputStream inputStream(IRubyObject in) {
         setBinmodeIfPossible(in);
-        if (in instanceof RubyIO) {
-            return ((RubyIO) in).getInStream();
-        }
         return new IOInputStream(in);
     }
 
-    private static void dumpToStream(IRubyObject object, OutputStream rawOutput, int depthLimit)
+    private static boolean dumpToStream(IRubyObject object, OutputStream rawOutput, int depthLimit)
         throws IOException
     {
         MarshalStream output = new MarshalStream(object.getRuntime(), rawOutput, depthLimit);
         output.dumpObject(object);
+        return output.isTainted();
     }
 }
