@@ -78,11 +78,6 @@ public class ConstNode extends Node implements INameNode {
         return EMPTY_LIST;
     }
     
-    @Override
-    public String toString() {
-        return "ConstNode [" + name + "]";
-    }
-
     public void setName(String name) {
         this.name = name;
     }
@@ -93,7 +88,7 @@ public class ConstNode extends Node implements INameNode {
 
         // We can callsite cache const_missing if we want
         return value != null ? value :
-            context.getRubyClass().callMethod(context, "const_missing", runtime.fastNewSymbol(name));
+            context.getCurrentScope().getStaticScope().getModule().callMethod(context, "const_missing", runtime.fastNewSymbol(name));
     }
 
     @Override
