@@ -35,6 +35,8 @@ package org.jruby.ast;
 import java.util.List;
 
 import org.jruby.Ruby;
+import org.jruby.RubyClass;
+import org.jruby.RubyClass.VariableAccessor;
 import org.jruby.ast.types.IArityNode;
 import org.jruby.ast.types.INameNode;
 import org.jruby.ast.visitor.NodeVisitor;
@@ -50,6 +52,7 @@ import org.jruby.runtime.builtin.IRubyObject;
  */
 public class InstVarNode extends Node implements IArityNode, INameNode {
     private String name;
+    private VariableAccessor accessor = VariableAccessor.DUMMY_ACCESSOR;
 
     public InstVarNode(ISourcePosition position, String name) {
         super(position);
@@ -93,10 +96,18 @@ public class InstVarNode extends Node implements IArityNode, INameNode {
     
     @Override
     public IRubyObject interpret(Ruby runtime, ThreadContext context, IRubyObject self, Block aBlock) {
-        IRubyObject variable = self.getInstanceVariables().fastGetInstanceVariable(name);
-   
-        if (variable != null) return variable;
-        
+        RubyClass cls = self.getMetaClass().getRealClass();
+        VariableAccessor localAccessor = accessor;
+        IRubyObject value;
+        if (localAccessor.getClassId() != cls.hashCode()) {
+            localAccessor = cls.getVariableAccessorForRead(name);
+            if (localAccessor == null) return runtime.getNil();
+            value = (IRubyObject)localAccessor.get(self);
+            accessor = localAccessor;
+        } else {
+            value = (IRubyObject)localAccessor.get(self);
+        }
+        if (value != null) return value;
         if (runtime.isVerbose()) warnAboutUninitializedIvar(runtime);
         return runtime.getNil();        
     }
@@ -110,10 +121,4 @@ public class InstVarNode extends Node implements IArityNode, INameNode {
     public String definition(Ruby runtime, ThreadContext context, IRubyObject self, Block aBlock) {
         return self.getInstanceVariables().fastHasInstanceVariable(name) ? "instance-variable" : null;
     }
-
-    @Override
-    public String toString() {
-        return "InstVarnode: " + name;
-
-    }
 }
