@@ -37,9 +37,10 @@ import org.jruby.Ruby;
 import org.jruby.RubyModule;
 import org.jruby.ast.visitor.NodeVisitor;
 import org.jruby.evaluator.ASTInterpreter;
-import org.jruby.javasupport.util.RuntimeHelpers;
 import org.jruby.lexer.yacc.ISourcePosition;
 import org.jruby.runtime.Block;
+import org.jruby.runtime.CallSite;
+import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 
@@ -49,6 +50,7 @@ import org.jruby.runtime.builtin.IRubyObject;
 public class SuperNode extends Node implements BlockAcceptingNode {
     private final Node argsNode;
     private Node iterNode;
+    private final CallSite callSite;
 
     public SuperNode(ISourcePosition position, Node argsNode) {
         this(position, argsNode, null);
@@ -58,6 +60,10 @@ public class SuperNode extends Node implements BlockAcceptingNode {
         super(position);
         this.argsNode = argsNode;
         this.iterNode = iterNode;
+        if (argsNode instanceof ArrayNode) {
+            ((ArrayNode)argsNode).setLightweight(true);
+        }
+        this.callSite = MethodIndex.getSuperCallSite();
     }
 
     public NodeType getNodeType() {
@@ -96,15 +102,13 @@ public class SuperNode extends Node implements BlockAcceptingNode {
 
     @Override
     public IRubyObject interpret(Ruby runtime, ThreadContext context, IRubyObject self, Block aBlock) {
-        RuntimeHelpers.checkSuperDisabledOrOutOfMethod(context);
-
         IRubyObject[] args = ASTInterpreter.setupArgs(runtime, context, argsNode, self, aBlock);
         Block block = ASTInterpreter.getBlock(runtime, context, self, aBlock, iterNode);
         
         // If no explicit block passed to super, then use the one passed in, unless it's explicitly cleared with nil
         if (iterNode == null && !block.isGiven()) block = aBlock;
-        
-        return RuntimeHelpers.invokeSuper(context, self, args, block);
+
+        return callSite.call(context, self, self, args, block);
     }
     
     @Override
