@@ -30,7 +30,6 @@ package org.jruby.ext.socket;
 import java.io.FileDescriptor;
 import java.io.IOException;
 import java.net.ConnectException;
-import java.net.Inet4Address;
 import java.net.InetAddress;
 import java.net.InetSocketAddress;
 import java.net.SocketException;
@@ -43,6 +42,7 @@ import org.jruby.Ruby;
 import org.jruby.RubyClass;
 import org.jruby.RubyFixnum;
 import org.jruby.RubyIO;
+import org.jruby.RubyModule;
 import org.jruby.RubyNumeric;
 import org.jruby.RubyString;
 import org.jruby.anno.JRubyClass;
@@ -100,9 +100,9 @@ public class RubyUDPSocket extends RubyIPSocket {
     }
 
     @JRubyMethod(visibility = Visibility.PRIVATE)
-    public IRubyObject initialize(IRubyObject protocol) {
+    public IRubyObject initialize(ThreadContext context, IRubyObject protocol) {
         // we basically ignore protocol. let someone report it...
-        return initialize();
+        return initialize(context);
     }
     
     @Deprecated
@@ -113,8 +113,18 @@ public class RubyUDPSocket extends RubyIPSocket {
     public IRubyObject bind(ThreadContext context, IRubyObject host, IRubyObject port) {
         InetSocketAddress addr = null;
         try {
-            if (host.isNil()) {
+            if (host.isNil()
+                || ((host instanceof RubyString)
+                && ((RubyString) host).isEmpty())) {
+                // host is nil or the empty string, bind to INADDR_ANY
                 addr = new InetSocketAddress(RubyNumeric.fix2int(port));
+            } else if (host instanceof RubyFixnum) {
+                // passing in something like INADDR_ANY
+                int intAddr = RubyNumeric.fix2int(host);
+                RubyModule socketMod = context.getRuntime().getModule("Socket");
+                if (intAddr == RubyNumeric.fix2int(socketMod.fastGetConstant("INADDR_ANY"))) {
+                    addr = new InetSocketAddress(InetAddress.getByName("0.0.0.0"), RubyNumeric.fix2int(port));
+                }
             } else {
                 addr = new InetSocketAddress(InetAddress.getByName(host.convertToString().toString()), RubyNumeric.fix2int(port));
             }
@@ -165,6 +175,12 @@ public class RubyUDPSocket extends RubyIPSocket {
             ((DatagramChannel) this.getChannel()).configureBlocking(false);
             context.getThread().select(this, SelectionKey.OP_READ);
             InetSocketAddress sender = (InetSocketAddress) ((DatagramChannel) this.getChannel()).receive(buf);
+
+            // see JRUBY-4678
+            if (sender == null) {
+                throw context.getRuntime().newErrnoECONNRESETError();
+            }
+
             IRubyObject addressArray = context.getRuntime().newArray(new IRubyObject[]{
                 context.getRuntime().newString("AF_INET"),
                 context.getRuntime().newFixnum(sender.getPort()),
@@ -185,13 +201,21 @@ public class RubyUDPSocket extends RubyIPSocket {
         try {
             int length = RubyNumeric.fix2int(args[0]);
             ByteBuffer buf = ByteBuffer.allocate(length);
+            ((DatagramChannel) this.getChannel()).configureBlocking(false);
             context.getThread().select(this, SelectionKey.OP_READ);
-            ((DatagramChannel) this.getChannel()).receive(buf);
+            InetSocketAddress sender = (InetSocketAddress) ((DatagramChannel) this.getChannel()).receive(buf);
+
+            // see JRUBY-4678
+            if (sender == null) {
+                throw context.getRuntime().newErrnoECONNRESETError();
+            }
+
             return context.getRuntime().newString(new ByteList(buf.array(), 0, buf.position()));
         } catch (IOException e) {
             throw sockerr(context.getRuntime(), "recv: name or service not known");
         }
     }
+
     @Deprecated
     public IRubyObject send(IRubyObject[] args) {
         return send(getRuntime().getCurrentContext(), args);
@@ -205,10 +229,21 @@ public class RubyUDPSocket extends RubyIPSocket {
                 RubyString nameStr = args[2].convertToString();
                 RubyString data = args[0].convertToString();
                 ByteBuffer buf = ByteBuffer.wrap(data.getBytes());
+                int port;
+                if (args[3] instanceof RubyString) {
+                    jnr.netdb.Service service = jnr.netdb.Service.getServiceByName(args[3].asJavaString(), "udp");
+                    if (service != null) {
+                        port = service.getPort();
+                    } else {
+                        port = (int)args[3].convertToInteger("to_i").getLongValue();
+                    }
+                } else {
+                    port = (int)args[3].convertToInteger().getLongValue();
+                }
 
                 InetAddress address = RubySocket.getRubyInetAddress(nameStr.getByteList());
                 InetSocketAddress addr =
-                        new InetSocketAddress(address, RubyNumeric.fix2int(args[3]));
+                        new InetSocketAddress(address, port);
                 written = ((DatagramChannel) this.getChannel()).send(buf, addr);
             } else {
                 RubyString data = args[0].convertToString();
