@@ -21,21 +21,29 @@ public class InstanceMethodInvoker extends MethodInvoker {
 
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args) {
-        createJavaMethods(context.getRuntime());
         JavaProxy proxy = castJavaProxy(self);
 
         int len = args.length;
         Object[] convertedArgs = new Object[len];
         JavaMethod method = (JavaMethod)findCallable(self, name, args, len);
-        for (int i = 0; i < len; i++) {
-            convertedArgs[i] = convertArg(context, args[i], method, i);
+        if (method.isVarArgs()) {
+            len = method.getParameterTypes().length - 1;
+            convertedArgs = new Object[len + 1];
+            for (int i = 0; i < len; i++) {
+                convertedArgs[i] = convertArg(args[i], method, i);
+            }
+            convertedArgs[len] = convertVarargs(args, method);
+        } else {
+            convertedArgs = new Object[len];
+            for (int i = 0; i < len; i++) {
+                convertedArgs[i] = convertArg(args[i], method, i);
+            }
         }
         return method.invokeDirect(proxy.getObject(), convertedArgs);
     }
 
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name) {
-        createJavaMethods(self.getRuntime());
         JavaProxy proxy = castJavaProxy(self);
         JavaMethod method = (JavaMethod)findCallableArityZero(self, name);
         return method.invokeDirect(proxy.getObject());
@@ -43,47 +51,46 @@ public class InstanceMethodInvoker extends MethodInvoker {
 
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0) {
-        createJavaMethods(self.getRuntime());
+        if (javaVarargsCallables != null) return call(context, self, clazz, name, new IRubyObject[] {arg0});
         JavaProxy proxy = castJavaProxy(self);
         JavaMethod method = (JavaMethod)findCallableArityOne(self, name, arg0);
-        Object cArg0 = convertArg(context, arg0, method, 0);
+        Object cArg0 = convertArg(arg0, method, 0);
         return method.invokeDirect(proxy.getObject(), cArg0);
     }
 
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0, IRubyObject arg1) {
-        createJavaMethods(self.getRuntime());
+        if (javaVarargsCallables != null) return call(context, self, clazz, name, new IRubyObject[] {arg0, arg1});
         JavaProxy proxy = castJavaProxy(self);
         JavaMethod method = (JavaMethod)findCallableArityTwo(self, name, arg0, arg1);
-        Object cArg0 = convertArg(context, arg0, method, 0);
-        Object cArg1 = convertArg(context, arg1, method, 1);
+        Object cArg0 = convertArg(arg0, method, 0);
+        Object cArg1 = convertArg(arg1, method, 1);
         return method.invokeDirect(proxy.getObject(), cArg0, cArg1);
     }
 
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0, IRubyObject arg1, IRubyObject arg2) {
-        createJavaMethods(self.getRuntime());
+        if (javaVarargsCallables != null) return call(context, self, clazz, name, new IRubyObject[] {arg0, arg1, arg2});
         JavaProxy proxy = castJavaProxy(self);
         JavaMethod method = (JavaMethod)findCallableArityThree(self, name, arg0, arg1, arg2);
-        Object cArg0 = convertArg(context, arg0, method, 0);
-        Object cArg1 = convertArg(context, arg1, method, 1);
-        Object cArg2 = convertArg(context, arg2, method, 2);
+        Object cArg0 = convertArg(arg0, method, 0);
+        Object cArg1 = convertArg(arg1, method, 1);
+        Object cArg2 = convertArg(arg2, method, 2);
         return method.invokeDirect(proxy.getObject(), cArg0, cArg1, cArg2);
     }
 
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args, Block block) {
         if (block.isGiven()) {
-            createJavaMethods(self.getRuntime());
             JavaProxy proxy = castJavaProxy(self);
             int len = args.length;
             // these extra arrays are really unfortunate; split some of these paths out to eliminate?
             Object[] convertedArgs = new Object[len + 1];
             IRubyObject[] intermediate = new IRubyObject[len + 1];
             System.arraycopy(args, 0, intermediate, 0, len);
-            intermediate[len] = RubyProc.newProc(self.getRuntime(), block, Block.Type.LAMBDA);
+            intermediate[len] = RubyProc.newProc(context.getRuntime(), block, Block.Type.LAMBDA);
             JavaMethod method = (JavaMethod)findCallable(self, name, intermediate, len + 1);
             for (int i = 0; i < len + 1; i++) {
-                convertedArgs[i] = convertArg(context, intermediate[i], method, i);
+                convertedArgs[i] = convertArg(intermediate[i], method, i);
             }
             return method.invokeDirect(proxy.getObject(), convertedArgs);
         } else {
@@ -94,11 +101,10 @@ public class InstanceMethodInvoker extends MethodInvoker {
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, Block block) {
         if (block.isGiven()) {
-            createJavaMethods(self.getRuntime());
             JavaProxy proxy = castJavaProxy(self);
-            RubyProc proc = RubyProc.newProc(self.getRuntime(), block, Block.Type.LAMBDA);
+            RubyProc proc = RubyProc.newProc(context.getRuntime(), block, Block.Type.LAMBDA);
             JavaMethod method = (JavaMethod)findCallableArityOne(self, name, proc);
-            Object cArg0 = convertArg(context, proc, method, 0);
+            Object cArg0 = convertArg(proc, method, 0);
             return method.invokeDirect(proxy.getObject(), cArg0);
         } else {
             return call(context, self, clazz, name);
@@ -108,12 +114,11 @@ public class InstanceMethodInvoker extends MethodInvoker {
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0, Block block) {
         if (block.isGiven()) {
-            createJavaMethods(self.getRuntime());
             JavaProxy proxy = castJavaProxy(self);
-            RubyProc proc = RubyProc.newProc(self.getRuntime(), block, Block.Type.LAMBDA);
+            RubyProc proc = RubyProc.newProc(context.getRuntime(), block, Block.Type.LAMBDA);
             JavaMethod method = (JavaMethod)findCallableArityTwo(self, name, arg0, proc);
-            Object cArg0 = convertArg(context, arg0, method, 0);
-            Object cArg1 = convertArg(context, proc, method, 1);
+            Object cArg0 = convertArg(arg0, method, 0);
+            Object cArg1 = convertArg(proc, method, 1);
             return method.invokeDirect(proxy.getObject(), cArg0, cArg1);
         } else {
             return call(context, self, clazz, name, arg0);
@@ -123,13 +128,12 @@ public class InstanceMethodInvoker extends MethodInvoker {
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0, IRubyObject arg1, Block block) {
         if (block.isGiven()) {
-            createJavaMethods(self.getRuntime());
             JavaProxy proxy = castJavaProxy(self);
-            RubyProc proc = RubyProc.newProc(self.getRuntime(), block, Block.Type.LAMBDA);
+            RubyProc proc = RubyProc.newProc(context.getRuntime(), block, Block.Type.LAMBDA);
             JavaMethod method = (JavaMethod)findCallableArityThree(self, name, arg0, arg1, proc);
-            Object cArg0 = convertArg(context, arg0, method, 0);
-            Object cArg1 = convertArg(context, arg1, method, 1);
-            Object cArg2 = convertArg(context, proc, method, 2);
+            Object cArg0 = convertArg(arg0, method, 0);
+            Object cArg1 = convertArg(arg1, method, 1);
+            Object cArg2 = convertArg(proc, method, 2);
             return method.invokeDirect(proxy.getObject(), cArg0, cArg1, cArg2);
         } else {
             return call(context, self, clazz, name, arg0, arg1);
@@ -139,14 +143,13 @@ public class InstanceMethodInvoker extends MethodInvoker {
     @Override
     public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject arg0, IRubyObject arg1, IRubyObject arg2, Block block) {
         if (block.isGiven()) {
-            createJavaMethods(self.getRuntime());
             JavaProxy proxy = castJavaProxy(self);
-            RubyProc proc = RubyProc.newProc(self.getRuntime(), block, Block.Type.LAMBDA);
+            RubyProc proc = RubyProc.newProc(context.getRuntime(), block, Block.Type.LAMBDA);
             JavaMethod method = (JavaMethod)findCallableArityFour(self, name, arg0, arg1, arg2, proc);
-            Object cArg0 = convertArg(context, arg0, method, 0);
-            Object cArg1 = convertArg(context, arg1, method, 1);
-            Object cArg2 = convertArg(context, arg2, method, 2);
-            Object cArg3 = convertArg(context, proc, method, 3);
+            Object cArg0 = convertArg(arg0, method, 0);
+            Object cArg1 = convertArg(arg1, method, 1);
+            Object cArg2 = convertArg(arg2, method, 2);
+            Object cArg3 = convertArg(proc, method, 3);
             return method.invokeDirect(proxy.getObject(), cArg0, cArg1, cArg2, cArg3);
         } else {
             return call(context, self, clazz, name, arg0, arg1, arg2);
