@@ -38,8 +38,6 @@ import org.jruby.java.addons.KernelJavaAddons;
 import java.io.IOException;
 import java.lang.reflect.Constructor;
 import java.util.ArrayList;
-import java.util.Arrays;
-import java.util.List;
 import java.util.HashMap;
 import java.util.Map;
 import java.util.regex.Matcher;
@@ -51,20 +49,23 @@ import java.lang.reflect.Modifier;
 import java.lang.reflect.Proxy;
 import java.math.BigDecimal;
 import java.math.BigInteger;
+import java.util.HashSet;
+import java.util.Set;
 import java.util.concurrent.ConcurrentHashMap;
 
 import org.jruby.MetaClass;
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
+import org.jruby.RubyBasicObject;
 import org.jruby.RubyClass;
 import org.jruby.RubyClassPathVariable;
 import org.jruby.RubyException;
+import org.jruby.RubyInstanceConfig;
 import org.jruby.RubyMethod;
 import org.jruby.RubyModule;
-import org.jruby.RubyProc;
+import org.jruby.RubyObject;
 import org.jruby.RubyString;
 import org.jruby.RubyUnboundMethod;
-import org.jruby.common.IRubyWarnings.ID;
 import org.jruby.exceptions.RaiseException;
 import org.jruby.javasupport.proxy.JavaProxyClass;
 import org.jruby.javasupport.proxy.JavaProxyConstructor;
@@ -80,23 +81,27 @@ import org.jruby.anno.JRubyMethod;
 import org.jruby.anno.JRubyModule;
 import org.jruby.internal.runtime.methods.JavaMethod.JavaMethodN;
 import org.jruby.internal.runtime.methods.JavaMethod.JavaMethodZero;
-import org.jruby.java.MiniJava;
 import org.jruby.java.addons.ArrayJavaAddons;
 import org.jruby.java.addons.IOJavaAddons;
 import org.jruby.java.addons.StringJavaAddons;
+import org.jruby.java.codegen.RealClassGenerator;
 import org.jruby.java.dispatch.CallableSelector;
 import org.jruby.java.invokers.InstanceMethodInvoker;
 import org.jruby.java.invokers.MethodInvoker;
 import org.jruby.java.invokers.StaticMethodInvoker;
 import org.jruby.java.proxies.ArrayJavaProxy;
 import org.jruby.java.proxies.ConcreteJavaProxy;
+import org.jruby.java.proxies.InterfaceJavaProxy;
 import org.jruby.java.proxies.JavaProxy;
-import org.jruby.runtime.callback.Callback;
+import org.jruby.java.proxies.RubyObjectHolderProxy;
 import org.jruby.util.CodegenUtils;
+import org.jruby.util.IdUtil;
 import org.jruby.util.SafePropertyAccessor;
 
 @JRubyModule(name = "Java")
 public class Java implements Library {
+    public static final boolean NEW_STYLE_EXTENSION = SafePropertyAccessor.getBoolean("jruby.ji.newStyleExtension", false);
+    public static final boolean OBJECT_PROXY_CACHE = SafePropertyAccessor.getBoolean("jruby.ji.objectProxyCache", true);
 
     public void load(Ruby runtime, boolean wrap) throws IOException {
         createJavaModule(runtime);
@@ -123,7 +128,7 @@ public class Java implements Library {
         // the proxy (wrapper) type hierarchy
         JavaProxy.createJavaProxy(context);
         ConcreteJavaProxy.createConcreteJavaProxy(context);
-        //InterfaceJavaProxy.createInterfaceJavaProxy(context);
+        InterfaceJavaProxy.createInterfaceJavaProxy(context);
         ArrayJavaProxy.createArrayJavaProxy(context);
 
         // also create the JavaProxy* classes
@@ -136,18 +141,6 @@ public class Java implements Library {
         
         javaUtils.defineAnnotatedMethods(JavaUtilities.class);
 
-        runtime.getJavaSupport().setConcreteProxyCallback(new Callback() {
-            public IRubyObject execute(IRubyObject recv, IRubyObject[] args, Block block) {
-                Arity.checkArgumentCount(recv.getRuntime(), args, 1, 1);
-                
-                return Java.concrete_proxy_inherited(recv, args[0]);
-            }
-
-            public Arity getArity() {
-                return Arity.ONE_ARGUMENT;
-            }
-        });
-
         JavaArrayUtilities.createJavaArrayUtilitiesModule(runtime);
         
         // Now attach Java-related extras to core classes
@@ -171,6 +164,25 @@ public class Java implements Library {
 
         return javaModule;
     }
+
+    public static class OldStyleExtensionInherited {
+        @JRubyMethod
+        public static IRubyObject inherited(IRubyObject recv, IRubyObject arg0) {
+            return Java.concrete_proxy_inherited(recv, arg0);
+        }
+    };
+
+    public static class NewStyleExtensionInherited {
+        @JRubyMethod
+        public static IRubyObject inherited(IRubyObject recv, IRubyObject arg0) {
+            if (!(arg0 instanceof RubyClass)) {
+                throw recv.getRuntime().newTypeError(arg0, recv.getRuntime().getClassClass());
+            }
+            
+            JavaInterfaceTemplate.addRealImplClassNew((RubyClass)arg0);
+            return recv.getRuntime().getNil();
+        }
+    };
     
     /**
      * This populates the master map from short-cut names to JavaClass instances for
@@ -296,10 +308,23 @@ public class Java implements Library {
             IRubyObject constant,
             IRubyObject javaClass,
             IRubyObject module) {
+        Ruby runtime = recv.getRuntime();
+
         if (!(module instanceof RubyModule)) {
-            throw recv.getRuntime().newTypeError(module, recv.getRuntime().getModule());
+            throw runtime.newTypeError(module, runtime.getModule());
         }
-        return ((RubyModule) module).const_set(constant, get_proxy_class(recv, javaClass));
+        IRubyObject proxyClass = get_proxy_class(recv, javaClass);
+        RubyModule m = (RubyModule)module;
+        String constName = constant.asJavaString();
+        IRubyObject existing = m.getConstantNoConstMissing(constName);
+
+        if (existing != null
+                && existing != RubyBasicObject.UNDEF
+                && existing != proxyClass) {
+            runtime.getWarnings().warn("replacing " + existing + " with " + proxyClass + " in constant '" + constName + " on class/module " + m);
+        }
+        
+        return ((RubyModule) module).setConstantQuiet(constant.asJavaString(), get_proxy_class(recv, javaClass));
     }
 
     public static IRubyObject get_java_class(IRubyObject recv, IRubyObject name) {
@@ -325,8 +350,12 @@ public class Java implements Library {
      */
     public static IRubyObject getInstance(Ruby runtime, Object rawJavaObject) {
         if (rawJavaObject != null) {
-            return runtime.getJavaSupport().getObjectProxyCache().getOrCreate(rawJavaObject,
-                    (RubyClass) getProxyClass(runtime, JavaClass.get(runtime, rawJavaObject.getClass())));
+            if (OBJECT_PROXY_CACHE) {
+                return runtime.getJavaSupport().getObjectProxyCache().getOrCreate(rawJavaObject,
+                        (RubyClass) getProxyClass(runtime, JavaClass.get(runtime, rawJavaObject.getClass())));
+            } else {
+                return allocateProxy(rawJavaObject, (RubyClass)getProxyClass(runtime, JavaClass.get(runtime, rawJavaObject.getClass())));
+            }
         }
         return runtime.getNil();
     }
@@ -372,6 +401,10 @@ public class Java implements Library {
         return getInterfaceModule(runtime, javaClass);
     }
 
+    public static RubyClass getProxyClassForObject(Ruby runtime, Object object) {
+        return (RubyClass)getProxyClass(runtime, JavaClass.get(runtime, object.getClass()));
+    }
+
     public static RubyModule getProxyClass(Ruby runtime, JavaClass javaClass) {
         RubyClass proxyClass;
         final Class<?> c = javaClass.javaClass();
@@ -400,8 +433,11 @@ public class Java implements Library {
                     proxyClass = createProxyClass(runtime,
                             runtime.getJavaSupport().getConcreteProxyClass(),
                             javaClass, true);
-                    proxyClass.getMetaClass().defineFastMethod("inherited",
-                            runtime.getJavaSupport().getConcreteProxyCallback());
+                    if (NEW_STYLE_EXTENSION) {
+                        proxyClass.getMetaClass().defineAnnotatedMethods(NewStyleExtensionInherited.class);
+                    } else {
+                        proxyClass.getMetaClass().defineAnnotatedMethods(OldStyleExtensionInherited.class);
+                    }
                     addToJavaPackageModule(proxyClass, javaClass);
 
                 } else {
@@ -453,10 +489,9 @@ public class Java implements Library {
 
     private static RubyClass createProxyClass(Ruby runtime, RubyClass baseType,
             JavaClass javaClass, boolean invokeInherited) {
-	// JRUBY-2938 the proxy class might already exist
-	RubyClass proxyClass = javaClass.getProxyClass();
-	if (proxyClass != null)
-	    return proxyClass;
+        // JRUBY-2938 the proxy class might already exist
+        RubyClass proxyClass = javaClass.getProxyClass();
+        if (proxyClass != null) return proxyClass;
 
         // this needs to be split, since conditional calling #inherited doesn't fit standard ruby semantics
         RubyClass.checkInheritable(baseType);
@@ -492,6 +527,69 @@ public class Java implements Library {
             return getRubyMethod(context, proxyClass, name, argTypesClasses);
         }
 
+        @JRubyMethod(backtrace = true, meta = true)
+        public static IRubyObject java_send(ThreadContext context, IRubyObject recv, IRubyObject rubyName) {
+            String name = rubyName.asJavaString();
+            Ruby runtime = context.getRuntime();
+
+            JavaMethod method = new JavaMethod(runtime, getMethodFromClass(runtime, recv, name));
+            return method.invokeStaticDirect();
+        }
+
+        @JRubyMethod(backtrace = true, meta = true)
+        public static IRubyObject java_send(ThreadContext context, IRubyObject recv, IRubyObject rubyName, IRubyObject argTypes) {
+            String name = rubyName.asJavaString();
+            RubyArray argTypesAry = argTypes.convertToArray();
+            Ruby runtime = context.getRuntime();
+
+            if (argTypesAry.size() != 0) {
+                Class[] argTypesClasses = (Class[]) argTypesAry.toArray(new Class[argTypesAry.size()]);
+                throw JavaMethod.newArgSizeMismatchError(runtime, argTypesClasses);
+            }
+
+            JavaMethod method = new JavaMethod(runtime, getMethodFromClass(runtime, recv, name));
+            return method.invokeStaticDirect();
+        }
+
+        @JRubyMethod(backtrace = true, meta = true)
+        public static IRubyObject java_send(ThreadContext context, IRubyObject recv, IRubyObject rubyName, IRubyObject argTypes, IRubyObject arg0) {
+            String name = rubyName.asJavaString();
+            RubyArray argTypesAry = argTypes.convertToArray();
+            Ruby runtime = context.getRuntime();
+
+            if (argTypesAry.size() != 1) {
+                throw JavaMethod.newArgSizeMismatchError(runtime, (Class) argTypesAry.eltInternal(0).toJava(Class.class));
+            }
+
+            Class argTypeClass = (Class) argTypesAry.eltInternal(0).toJava(Class.class);
+
+            JavaMethod method = new JavaMethod(runtime, getMethodFromClass(runtime, recv, name, argTypeClass));
+            return method.invokeStaticDirect(arg0.toJava(argTypeClass));
+        }
+
+        @JRubyMethod(required = 4, rest = true, backtrace = true, meta = true)
+        public static IRubyObject java_send(ThreadContext context, IRubyObject recv, IRubyObject[] args) {
+            Ruby runtime = context.getRuntime();
+
+            String name = args[0].asJavaString();
+            RubyArray argTypesAry = args[1].convertToArray();
+            int argsLen = args.length - 2;
+
+            if (argTypesAry.size() != argsLen) {
+                throw JavaMethod.newArgSizeMismatchError(runtime, (Class[]) argTypesAry.toArray(new Class[argTypesAry.size()]));
+            }
+
+            Class[] argTypesClasses = (Class[]) argTypesAry.toArray(new Class[argsLen]);
+
+            Object[] argsAry = new Object[argsLen];
+            for (int i = 0; i < argsLen; i++) {
+                argsAry[i] = args[i + 2].toJava(argTypesClasses[i]);
+            }
+
+            JavaMethod method = new JavaMethod(runtime, getMethodFromClass(runtime, recv, name, argTypesClasses));
+            return method.invokeStaticDirect(argsAry);
+        }
+
         @JRubyMethod(backtrace = true, meta = true, visibility = Visibility.PRIVATE)
         public static IRubyObject java_alias(ThreadContext context, IRubyObject proxyClass, IRubyObject newName, IRubyObject rubyName) {
             return java_alias(context, proxyClass, newName, rubyName, context.getRuntime().newEmptyArray());
@@ -512,9 +610,7 @@ public class Java implements Library {
                 throw runtime.newTypeError(proxyClass, runtime.getModule());
             }
 
-            Class jclass = (Class)((JavaClass)proxyClass.callMethod(context, "java_class")).getValue();
-
-            Method method = getMethodFromClass(runtime, jclass, name, argTypesClasses);
+            Method method = getMethodFromClass(runtime, proxyClass, name, argTypesClasses);
             MethodInvoker invoker = getMethodInvokerForMethod(rubyClass, method);
 
             if (Modifier.isStatic(method.getModifiers())) {
@@ -538,30 +634,27 @@ public class Java implements Library {
             throw runtime.newTypeError(proxyClass, runtime.getModule());
         }
 
-        Class jclass = (Class)((JavaClass)proxyClass.callMethod(context, "java_class")).getValue();
+        Method jmethod = getMethodFromClass(runtime, proxyClass, name, argTypesClasses);
         String prettyName = name + CodegenUtils.prettyParams(argTypesClasses);
         
-        try {
-            Method jmethod = jclass.getMethod(name, argTypesClasses);
-            MethodInvoker invoker;
-            if (Modifier.isStatic(jmethod.getModifiers())) {
-                invoker = new StaticMethodInvoker(rubyClass, jmethod);
-                return RubyMethod.newMethod(rubyClass, prettyName, rubyClass, name, invoker, proxyClass);
-            } else {
-                invoker = new InstanceMethodInvoker(rubyClass, jmethod);
-                return RubyUnboundMethod.newUnboundMethod(rubyClass, prettyName, rubyClass, name, invoker);
-            }
-        } catch (NoSuchMethodException nsme) {
-            String errorName = jclass.getName() + "." + prettyName;
-            throw runtime.newNameError("Java method not found: " + errorName, name);
+        if (Modifier.isStatic(jmethod.getModifiers())) {
+            MethodInvoker invoker = new StaticMethodInvoker(rubyClass, jmethod);
+            return RubyMethod.newMethod(rubyClass, prettyName, rubyClass, name, invoker, proxyClass);
+        } else {
+            MethodInvoker invoker = new InstanceMethodInvoker(rubyClass, jmethod);
+            return RubyUnboundMethod.newUnboundMethod(rubyClass, prettyName, rubyClass, name, invoker);
         }
     }
 
-    public static Method getMethodFromClass(Ruby runtime, Class jclass, String name, Class... argTypes) {
+    public static Method getMethodFromClass(Ruby runtime, IRubyObject proxyClass, String name, Class... argTypes) {
+        Class jclass = (Class)((JavaClass)proxyClass.callMethod(runtime.getCurrentContext(), "java_class")).getValue();
+
         try {
             return jclass.getMethod(name, argTypes);
         } catch (NoSuchMethodException nsme) {
-            throw JavaMethod.newMethodNotFoundError(runtime, jclass, name + CodegenUtils.prettyParams(argTypes), name);
+            String prettyName = name + CodegenUtils.prettyParams(argTypes);
+            String errorName = jclass.getName() + "." + prettyName;
+            throw runtime.newNameError("Java method not found: " + errorName, name);
         }
     }
 
@@ -625,7 +718,6 @@ public class Java implements Library {
                     throw context.getRuntime().newArgumentError("wrong # of arguments for constructor");
                 }
                 JavaProxyConstructor matching = (JavaProxyConstructor)CallableSelector.matchingCallableArityN(
-                        self,
                         methodCache,
                         forArity.toArray(new JavaProxyConstructor[forArity.size()]), args, args.length);
                 Object[] newArgs = new Object[args.length];
@@ -644,23 +736,34 @@ public class Java implements Library {
     // package scheme 2: separate module for each full package name, constructed 
     // from the camel-cased package segments: Java::JavaLang::Object, 
     private static void addToJavaPackageModule(RubyModule proxyClass, JavaClass javaClass) {
+        Ruby runtime = proxyClass.getRuntime();
         Class<?> clazz = javaClass.javaClass();
         String fullName;
         if ((fullName = clazz.getName()) == null) {
             return;
         }
         int endPackage = fullName.lastIndexOf('.');
-        // we'll only map conventional class names to modules 
-        if (fullName.indexOf('$') != -1 || !Character.isUpperCase(fullName.charAt(endPackage + 1))) {
-            return;
+        RubyModule parentModule;
+        String className;
+
+        // inner classes must be nested
+        if (fullName.indexOf('$') != -1) {
+            IRubyObject declClass = javaClass.declaring_class();
+            if (declClass.isNil()) {
+                // no containing class for a $ class; treat it as internal and don't define a constant
+                return;
+            }
+            parentModule = getProxyClass(runtime, (JavaClass)declClass);
+            className = clazz.getSimpleName();
+        } else {
+            String packageString = endPackage < 0 ? "" : fullName.substring(0, endPackage);
+            parentModule = getJavaPackageModule(runtime, packageString);
+            className = parentModule == null ? fullName : fullName.substring(endPackage + 1);
         }
-        Ruby runtime = proxyClass.getRuntime();
-        String packageString = endPackage < 0 ? "" : fullName.substring(0, endPackage);
-        RubyModule packageModule = getJavaPackageModule(runtime, packageString);
-        if (packageModule != null) {
-            String className = fullName.substring(endPackage + 1);
-            if (packageModule.getConstantAt(className) == null) {
-                packageModule.const_set(runtime.newSymbol(className), proxyClass);
+        
+        if (parentModule != null && IdUtil.isConstant(className)) {
+            if (parentModule.getConstantAt(className) == null) {
+                parentModule.setConstant(className, proxyClass);
             }
         }
     }
@@ -767,39 +870,24 @@ public class Java implements Library {
             } catch (Exception e) { /* expected */ }
 
             // Haven't found a class, continue on as though it were a package
-            RubyModule packageModule;
+            final RubyModule packageModule = getJavaPackageModule(runtime, fullName);
             // TODO: decompose getJavaPackageModule so we don't parse fullName
-            if ((packageModule = getJavaPackageModule(runtime, fullName)) == null) {
+            if (packageModule == null) {
                 return null;
             }
+
+            // save package in singletonized parent, so we don't come back here
+            memoizePackageOrClass(parentPackage, name, packageModule);
             
-            // save package module as ivar in parent, and add method to parent so
-            // we don't have to come back here.
-            final String ivarName = ("@__pkg__" + name).intern();
-            parentPackage.fastSetInstanceVariable(ivarName, packageModule);
-            RubyClass singleton = parentPackage.getSingletonClass();
-            singleton.addMethod(name, new org.jruby.internal.runtime.methods.JavaMethod(singleton, Visibility.PUBLIC) {
-
-                public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args, Block block) {
-                    if (args.length != 0) {
-                        Arity.raiseArgumentError(runtime, args.length, 0, 0);
-                    }
-                    IRubyObject variable;
-                    if ((variable = ((RubyModule) self).fastGetInstanceVariable(ivarName)) != null) {
-                        return variable;
-                    }
-                    return runtime.getNil();
-                }
-
-                @Override
-                public Arity getArity() {
-                    return Arity.noArguments();
-                }
-            });
             return packageModule;
         } else {
             // upper case name, so most likely a class
-            return getProxyClass(runtime, JavaClass.forNameVerbose(runtime, fullName));
+            final RubyModule javaModule = getProxyClass(runtime, JavaClass.forNameVerbose(runtime, fullName));
+
+            // save class in singletonized parent, so we don't come back here
+            memoizePackageOrClass(parentPackage, name, javaModule);
+
+            return javaModule;
 
         // FIXME: we should also support orgs that use capitalized package
         // names (including, embarrassingly, the one I work for), but this
@@ -852,42 +940,23 @@ public class Java implements Library {
 
             // TODO: check for Java reserved names and raise exception if encountered
 
-            RubyModule packageModule;
+            final RubyModule packageModule = getJavaPackageModule(runtime, name);
             // TODO: decompose getJavaPackageModule so we don't parse fullName
-            if ((packageModule = getJavaPackageModule(runtime, name)) == null) {
+            if (packageModule == null) {
                 return null;
             }
             RubyModule javaModule = runtime.getJavaSupport().getJavaModule();
             if (javaModule.getMetaClass().isMethodBound(name, false)) {
                 return packageModule;
-            // save package module as ivar in parent, and add method to parent so
-            // we don't have to come back here.
             }
-            final String ivarName = ("@__pkg__" + name).intern();
-            javaModule.fastSetInstanceVariable(ivarName, packageModule);
-            RubyClass singleton = javaModule.getSingletonClass();
-            singleton.addMethod(name, new org.jruby.internal.runtime.methods.JavaMethod(singleton, Visibility.PUBLIC) {
 
-                public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args, Block block) {
-                    if (args.length != 0) {
-                        Arity.raiseArgumentError(runtime, args.length, 0, 0);
-                    }
-                    IRubyObject variable;
-                    if ((variable = ((RubyModule) self).fastGetInstanceVariable(ivarName)) != null) {
-                        return variable;
-                    }
-                    return runtime.getNil();
-                }
-
-                @Override
-                public Arity getArity() {
-                    return Arity.noArguments();
-                }
-            });
+            memoizePackageOrClass(javaModule, name, packageModule);
+            
             return packageModule;
         } else {
+            RubyModule javaModule = null;
             try {
-                return getProxyClass(runtime, JavaClass.forNameQuiet(runtime, name));
+                javaModule = getProxyClass(runtime, JavaClass.forNameQuiet(runtime, name));
             } catch (RaiseException re) { /* not a class */
                 RubyException rubyEx = re.getException();
                 if (rubyEx.kind_of_p(context, runtime.getStandardError()).isTrue()) {
@@ -899,10 +968,42 @@ public class Java implements Library {
             // TODO: top-level upper-case package was supported in the previous (Ruby-based)
             // implementation, so leaving as is.  see note at #getProxyOrPackageUnderPackage
             // re: future approach below the top-level.
-            return getPackageModule(runtime, name);
+            if (javaModule == null) {
+                javaModule = getPackageModule(runtime, name);
+            }
+
+            memoizePackageOrClass(runtime.getJavaSupport().getJavaModule(), name, javaModule);
+
+            return javaModule;
         }
     }
 
+    private static void memoizePackageOrClass(final RubyModule parentPackage, final String name, final IRubyObject value) {
+        RubyClass singleton = parentPackage.getSingletonClass();
+        singleton.addMethod(name, new org.jruby.internal.runtime.methods.JavaMethod(singleton, Visibility.PUBLIC) {
+            public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name, IRubyObject[] args, Block block) {
+                if (args.length != 0) {
+                    throw context.getRuntime().newArgumentError(
+                            "Java package `"
+                            + parentPackage.callMethod("package_name")
+                            + "' does not have a method `"
+                            + name
+                            + "'");
+                }
+                return call(context, self, clazz, name);
+            }
+
+            public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name) {
+                return value;
+            }
+
+            @Override
+            public Arity getArity() {
+                return Arity.noArguments();
+            }
+        });
+    }
+
     public static IRubyObject get_top_level_proxy_or_package(ThreadContext context, IRubyObject recv, IRubyObject sym) {
         Ruby runtime = context.getRuntime();
         RubyModule result = getTopLevelProxyOrPackage(context, runtime, sym.asJavaString());
@@ -945,62 +1046,6 @@ public class Java implements Library {
         return JavaUtil.java_to_primitive(recv, object, unusedBlock);
     }
 
-    @JRubyMethod(required = 1, rest = true, frame = true, module = true, visibility = Visibility.PRIVATE)
-    @Deprecated
-    public static IRubyObject new_proxy_instance(final IRubyObject recv, IRubyObject[] args, Block block) {
-        int size = Arity.checkArgumentCount(recv.getRuntime(), args, 1, -1) - 1;
-        final RubyProc proc;
-
-        // Is there a supplied proc argument or do we assume a block was supplied
-        if (args[size] instanceof RubyProc) {
-            proc = (RubyProc) args[size];
-        } else {
-            proc = recv.getRuntime().newProc(Block.Type.PROC, block);
-            size++;
-        }
-
-        // Create list of interfaces to proxy (and make sure they really are interfaces)
-        Class[] interfaces = new Class[size];
-        for (int i = 0; i < size; i++) {
-            if (!(args[i] instanceof JavaClass) || !((JavaClass) args[i]).interface_p().isTrue()) {
-                throw recv.getRuntime().newArgumentError("Java interface expected. got: " + args[i]);
-            }
-            interfaces[i] = ((JavaClass) args[i]).javaClass();
-        }
-
-        return JavaObject.wrap(recv.getRuntime(), Proxy.newProxyInstance(recv.getRuntime().getJRubyClassLoader(), interfaces, new InvocationHandler() {
-
-            private Map parameterTypeCache = new ConcurrentHashMap();
-
-            public Object invoke(Object proxy, Method method, Object[] nargs) throws Throwable {
-                Class[] parameterTypes = (Class[]) parameterTypeCache.get(method);
-                if (parameterTypes == null) {
-                    parameterTypes = method.getParameterTypes();
-                    parameterTypeCache.put(method, parameterTypes);
-                }
-                int methodArgsLength = parameterTypes.length;
-                String methodName = method.getName();
-
-                if (methodName.equals("toString") && methodArgsLength == 0) {
-                    return proxy.getClass().getName();
-                } else if (methodName.equals("hashCode") && methodArgsLength == 0) {
-                    return new Integer(proxy.getClass().hashCode());
-                } else if (methodName.equals("equals") && methodArgsLength == 1 && parameterTypes[0].equals(Object.class)) {
-                    return Boolean.valueOf(proxy == nargs[0]);
-                }
-                Ruby runtime = recv.getRuntime();
-                int length = nargs == null ? 0 : nargs.length;
-                IRubyObject[] rubyArgs = new IRubyObject[length + 2];
-                rubyArgs[0] = JavaUtil.convertJavaToRuby(runtime, proxy);
-                rubyArgs[1] = new JavaMethod(runtime, method);
-                for (int i = 0; i < length; i++) {
-                    rubyArgs[i + 2] = JavaUtil.convertJavaToRuby(runtime, nargs[i]);
-                }
-                return proc.call(runtime.getCurrentContext(), rubyArgs).toJava(method.getReturnType());
-            }
-        }));
-    }
-
     @JRubyMethod(required = 2, frame = true, module = true, visibility = Visibility.PRIVATE)
     public static IRubyObject new_proxy_instance2(IRubyObject recv, final IRubyObject wrapper, IRubyObject ifcs, Block block) {
         IRubyObject[] javaClasses = ((RubyArray)ifcs).toJavaArray();
@@ -1016,9 +1061,20 @@ public class Java implements Library {
             interfaces[i] = ((JavaClass) javaClasses[i]).javaClass();
         }
 
+        return newInterfaceImpl(wrapper, interfaces);
+    }
+
+    public static IRubyObject newInterfaceImpl(final IRubyObject wrapper, Class[] interfaces) {
+        final Ruby runtime = wrapper.getRuntime();
+
+        Class[] tmp_interfaces = interfaces;
+        interfaces = new Class[tmp_interfaces.length + 1];
+        System.arraycopy(tmp_interfaces, 0, interfaces, 0, tmp_interfaces.length);
+        interfaces[tmp_interfaces.length] = RubyObjectHolderProxy.class;
+
         // hashcode is a combination of the interfaces and the Ruby class we're using
         // to implement them
-        if (!SafePropertyAccessor.getBoolean("jruby.interfaces.useProxy")) {
+        if (!RubyInstanceConfig.INTERFACES_USE_PROXY) {
             int interfacesHashCode = interfacesHashCode(interfaces);
             // if it's a singleton class and the real class is proc, we're doing closure conversion
             // so just use Proc's hashcode
@@ -1033,12 +1089,12 @@ public class Java implements Library {
             try {
                 proxyImplClass = Class.forName(implClassName, true, runtime.getJRubyClassLoader());
             } catch (ClassNotFoundException cnfe) {
-                proxyImplClass = MiniJava.createOldStyleImplClass(interfaces, wrapper.getMetaClass(), runtime, implClassName);
+                proxyImplClass = RealClassGenerator.createOldStyleImplClass(interfaces, wrapper.getMetaClass(), runtime, implClassName);
             }
 
             try {
                 Constructor proxyConstructor = proxyImplClass.getConstructor(IRubyObject.class);
-                return JavaObject.wrap(recv.getRuntime(), proxyConstructor.newInstance(wrapper));
+                return JavaObject.wrap(runtime, proxyConstructor.newInstance(wrapper));
             } catch (NoSuchMethodException nsme) {
                 throw runtime.newTypeError("Exception instantiating generated interface impl:\n" + nsme);
             } catch (InvocationTargetException ite) {
@@ -1049,7 +1105,7 @@ public class Java implements Library {
                 throw runtime.newTypeError("Exception instantiating generated interface impl:\n" + iae);
             }
         } else {
-            return JavaObject.wrap(recv.getRuntime(), Proxy.newProxyInstance(recv.getRuntime().getJRubyClassLoader(), interfaces, new InvocationHandler() {
+            Object proxyObject = Proxy.newProxyInstance(runtime.getJRubyClassLoader(), interfaces, new InvocationHandler() {
                 private Map parameterTypeCache = new ConcurrentHashMap();
 
                 public Object invoke(Object proxy, Method method, Object[] nargs) throws Throwable {
@@ -1060,7 +1116,7 @@ public class Java implements Library {
                     if (methodName == "toString" && length == 0) {
                         return proxy.getClass().getName();
                     } else if (methodName == "hashCode" && length == 0) {
-                        return new Integer(proxy.getClass().hashCode());
+                        return Integer.valueOf(proxy.getClass().hashCode());
                     } else if (methodName == "equals" && length == 1) {
                         Class[] parameterTypes = (Class[]) parameterTypeCache.get(method);
                         if (parameterTypes == null) {
@@ -1070,6 +1126,8 @@ public class Java implements Library {
                         if (parameterTypes[0].equals(Object.class)) {
                             return Boolean.valueOf(proxy == nargs[0]);
                         }
+                    } else if (methodName == "__ruby_object" && length == 0) {
+                        return wrapper;
                     }
 
                     IRubyObject[] rubyArgs = JavaUtil.convertJavaArrayToRuby(runtime, nargs);
@@ -1077,9 +1135,109 @@ public class Java implements Library {
                         return RuntimeHelpers.invoke(runtime.getCurrentContext(), wrapper, methodName, rubyArgs).toJava(method.getReturnType());
                     } catch (RuntimeException e) { e.printStackTrace(); throw e; }
                 }
-            }));
+            });
+            return JavaObject.wrap(runtime, proxyObject);
         }
     }
+
+    public static Class generateRealClass(final RubyClass clazz) {
+        final Ruby runtime = clazz.getRuntime();
+        final Class[] interfaces = getInterfacesFromRubyClass(clazz);
+
+        // hashcode is a combination of the interfaces and the Ruby class we're using
+        // to implement them
+        int interfacesHashCode = interfacesHashCode(interfaces);
+        // normal new class implementing interfaces
+        interfacesHashCode = 31 * interfacesHashCode + clazz.hashCode();
+        
+        String implClassName;
+        if (clazz.getBaseName() == null) {
+            // no-name class, generate a bogus name for it
+            implClassName = "anon_class" + Math.abs(System.identityHashCode(clazz)) + "_" + Math.abs(interfacesHashCode);
+        } else {
+            implClassName = clazz.getName().replaceAll("::", "\\$\\$") + "_" + Math.abs(interfacesHashCode);
+        }
+        Class proxyImplClass;
+        try {
+            proxyImplClass = Class.forName(implClassName, true, runtime.getJRubyClassLoader());
+        } catch (ClassNotFoundException cnfe) {
+            // try to use super's reified class; otherwise, RubyObject (for now)
+            Class superClass = clazz.getSuperClass().getRealClass().getReifiedClass();
+            if (superClass == null) {
+                superClass = RubyObject.class;
+            }
+            proxyImplClass = RealClassGenerator.createRealImplClass(superClass, interfaces, clazz, runtime, implClassName);
+            
+            // add a default initialize if one does not already exist and this is a Java-hierarchy class
+            if (NEW_STYLE_EXTENSION &&
+                    !(RubyBasicObject.class.isAssignableFrom(proxyImplClass) || clazz.getMethods().containsKey("initialize"))
+                    ) {
+                clazz.addMethod("initialize", new JavaMethodZero(clazz, Visibility.PUBLIC) {
+                    @Override
+                    public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name) {
+                        return context.getRuntime().getNil();
+                    }
+                });
+            }
+        }
+        clazz.setReifiedClass(proxyImplClass);
+
+        return proxyImplClass;
+    }
+
+    public static Constructor getRealClassConstructor(Ruby runtime, Class proxyImplClass) {
+        try {
+            return proxyImplClass.getConstructor(Ruby.class, RubyClass.class);
+        } catch (NoSuchMethodException nsme) {
+            throw runtime.newTypeError("Exception instantiating generated interface impl:\n" + nsme);
+        }
+    }
+
+    public static IRubyObject constructProxy(Ruby runtime, Constructor proxyConstructor, RubyClass clazz) {
+        try {
+            return (IRubyObject)proxyConstructor.newInstance(runtime, clazz);
+        } catch (InvocationTargetException ite) {
+            ite.printStackTrace();
+            throw runtime.newTypeError("Exception instantiating generated interface impl:\n" + ite);
+        } catch (InstantiationException ie) {
+            throw runtime.newTypeError("Exception instantiating generated interface impl:\n" + ie);
+        } catch (IllegalAccessException iae) {
+            throw runtime.newTypeError("Exception instantiating generated interface impl:\n" + iae);
+        }
+    }
+    
+    public static IRubyObject allocateProxy(Object javaObject, RubyClass clazz) {
+        IRubyObject proxy = clazz.allocate();
+        if (proxy instanceof JavaProxy) {
+            ((JavaProxy)proxy).setObject(javaObject);
+        } else {
+            JavaObject wrappedObject = JavaObject.wrap(clazz.getRuntime(), javaObject);
+            proxy.dataWrapStruct(wrappedObject);
+        }
+
+        return proxy;
+    }
+
+    public static IRubyObject wrapJavaObject(Ruby runtime, Object object) {
+        return allocateProxy(object, getProxyClassForObject(runtime, object));
+    }
+
+    public static Class[] getInterfacesFromRubyClass(RubyClass klass) {
+        Set<Class> interfaces = new HashSet<Class>();
+        // walk all superclasses aggregating interfaces
+        while (klass != null) {
+            IRubyObject maybeInterfaces = klass.getInstanceVariables().getInstanceVariable("@java_interfaces");
+            if (maybeInterfaces instanceof RubyArray) {
+                RubyArray moreInterfaces = (RubyArray)maybeInterfaces;
+                if (!moreInterfaces.isFrozen()) moreInterfaces.setFrozen(true);
+
+                interfaces.addAll(moreInterfaces);
+            }
+            klass = klass.getSuperClass();
+        }
+
+        return interfaces.toArray(new Class[interfaces.size()]);
+    }
     
     private static int interfacesHashCode(Class[] a) {
         if (a == null)
