@@ -49,6 +49,9 @@ import java.lang.reflect.Constructor;
 import java.lang.reflect.Field;
 import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
+import java.lang.reflect.ReflectPermission;
+import java.security.AccessControlException;
+import java.security.AccessController;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collections;
@@ -70,9 +73,11 @@ import org.jruby.anno.JRubyClass;
 import org.jruby.common.IRubyWarnings.ID;
 import org.jruby.exceptions.RaiseException;
 import org.jruby.internal.runtime.methods.DynamicMethod;
+import org.jruby.internal.runtime.methods.JavaMethod.JavaMethodZero;
 import org.jruby.java.addons.ArrayJavaAddons;
 import org.jruby.java.proxies.ArrayJavaProxy;
 import org.jruby.java.invokers.ConstructorInvoker;
+import org.jruby.java.proxies.ConcreteJavaProxy;
 import org.jruby.javasupport.util.RuntimeHelpers;
 import org.jruby.runtime.Arity;
 import org.jruby.runtime.Block;
@@ -82,14 +87,40 @@ import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.callback.Callback;
 import org.jruby.util.ByteList;
-import org.jruby.util.CodegenUtils;
 import org.jruby.util.IdUtil;
+import org.jruby.util.SafePropertyAccessor;
 
 
 @JRubyClass(name="Java::JavaClass", parent="Java::JavaObject")
 public class JavaClass extends JavaObject {
     public static final String METHOD_MANGLE = "__method";
 
+    public static final boolean CAN_SET_ACCESSIBLE;
+
+    static {
+        boolean canSetAccessible = false;
+
+        try {
+            AccessController.checkPermission(new ReflectPermission("suppressAccessChecks"));
+            canSetAccessible = true;
+        } catch (Throwable t) {
+            // added this so if things are weird in the future we can debug without
+            // spinning a new binary
+            if (SafePropertyAccessor.getBoolean("jruby.ji.logCanSetAccessible")) {
+                t.printStackTrace();
+            }
+            
+            // assume any exception means we can't suppress access checks
+            canSetAccessible = false;
+        }
+
+        CAN_SET_ACCESSIBLE = canSetAccessible;
+    }
+    
+    // An object that's never wrapped, so we can safely compare it with a
+    // wrapped null and not treat
+    private static final Object NEVER = new Object();
+    
     /**
      * Assigned names only override based priority of an assigned type, the type must be less than
      * or equal to the assigned type. For example, field name (FIELD) in a subclass will override
@@ -134,22 +165,21 @@ public class JavaClass extends JavaObject {
     static {
         RESERVED_NAMES.put("__id__", new AssignedName("__id__", Priority.RESERVED));
         RESERVED_NAMES.put("__send__", new AssignedName("__send__", Priority.RESERVED));
-        RESERVED_NAMES.put("class", new AssignedName("class", Priority.RESERVED));
-        RESERVED_NAMES.put("initialize", new AssignedName("initialize", Priority.RESERVED));
-        RESERVED_NAMES.put("object_id", new AssignedName("object_id", Priority.RESERVED));
         RESERVED_NAMES.put("private", new AssignedName("private", Priority.RESERVED));
         RESERVED_NAMES.put("protected", new AssignedName("protected", Priority.RESERVED));
         RESERVED_NAMES.put("public", new AssignedName("public", Priority.RESERVED));
-
-        // weakly reserved names
-        RESERVED_NAMES.put("id", new AssignedName("id", Priority.WEAKLY_RESERVED));
     }
     private static final Map<String, AssignedName> STATIC_RESERVED_NAMES = new HashMap<String, AssignedName>(RESERVED_NAMES);
     static {
         STATIC_RESERVED_NAMES.put("new", new AssignedName("new", Priority.RESERVED));
-        STATIC_RESERVED_NAMES.put("inspect", new AssignedName("inspect", Priority.RESERVED));
     }
     private static final Map<String, AssignedName> INSTANCE_RESERVED_NAMES = new HashMap<String, AssignedName>(RESERVED_NAMES);
+    static {
+        // only possible for "getClass" to be an instance method in Java
+        INSTANCE_RESERVED_NAMES.put("class", new AssignedName("class", Priority.RESERVED));
+        // "initialize" has meaning only for an instance (as opposed to a class)
+        INSTANCE_RESERVED_NAMES.put("initialize", new AssignedName("initialize", Priority.RESERVED));
+    }
 
     private static abstract class NamedInstaller {
         static final int STATIC_FIELD = 1;
@@ -250,11 +280,9 @@ public class JavaClass extends JavaObject {
             if (methods == null) {
                 methods = new ArrayList<Method>(4);
             }
-            if (!Ruby.isSecurityRestricted()) {
-                method.setAccessible(true);
-            }
             methods.add(method);
-            haveLocalMethod |= javaClass == method.getDeclaringClass();
+            haveLocalMethod |= javaClass == method.getDeclaringClass() ||
+                    method.getDeclaringClass().isInterface();
         }
 
         // called only by initializing thread; no synchronization required
@@ -287,7 +315,9 @@ public class JavaClass extends JavaObject {
                 constructors = new ArrayList<Constructor>(4);
             }
             if (!Ruby.isSecurityRestricted()) {
-                ctor.setAccessible(true);
+                try {
+                    ctor.setAccessible(true);
+                } catch(SecurityException e) {}
             }
             constructors.add(ctor);
             haveLocalConstructor |= javaClass == ctor.getDeclaringClass();
@@ -353,14 +383,10 @@ public class JavaClass extends JavaObject {
             if (proxy.getConstantAt(field.getName()) == null) {
                 // TODO: catch exception if constant is already set by other
                 // thread
-                if (!Ruby.isSecurityRestricted()) {
-                    field.setAccessible(true);
-                }
                 try {
                     proxy.setConstant(field.getName(), JavaUtil.convertJavaToUsableRubyObject(proxy.getRuntime(), field.get(null)));
                 } catch (IllegalAccessException iae) {
-                    throw proxy.getRuntime().newTypeError(
-                                        "illegal access on setting variable: " + iae.getMessage());
+                    // if we can't read it, we don't set it
                 }
             }
         }
@@ -398,6 +424,8 @@ public class JavaClass extends JavaObject {
     private RubyClass unfinishedProxyClass;
     
     private final ReentrantLock proxyLock = new ReentrantLock();
+
+    private final Initializer initializer;
     
     public RubyModule getProxyModule() {
         // allow proxy to be read without synchronization. if proxy
@@ -447,10 +475,11 @@ public class JavaClass extends JavaObject {
     private JavaClass(Ruby runtime, Class<?> javaClass) {
         super(runtime, (RubyClass) runtime.getJavaSupport().getJavaClassClass(), javaClass);
         if (javaClass.isInterface()) {
-            initializeInterface(javaClass);
+            initializer = new InterfaceInitializer(javaClass);
         } else if (!(javaClass.isArray() || javaClass.isPrimitive())) {
-            // TODO: public only?
-            initializeClass(javaClass);
+            initializer = new ClassInitializer(javaClass);
+        } else {
+            initializer = DUMMY_INITIALIZER;
         }
     }
     
@@ -459,64 +488,108 @@ public class JavaClass extends JavaObject {
         return other instanceof JavaClass &&
             this.getValue() == ((JavaClass)other).getValue();
     }
-    
-    private void initializeInterface(Class<?> javaClass) {
-        Map<String, AssignedName> staticNames  = new HashMap<String, AssignedName>(STATIC_RESERVED_NAMES);
-        List<ConstantField> constants = new ArrayList<ConstantField>(); 
-        Map<String, NamedInstaller> staticCallbacks = new HashMap<String, NamedInstaller>();
-        Field[] fields = getDeclaredFields(javaClass); 
 
-        for (int i = fields.length; --i >= 0; ) {
-            Field field = fields[i];
-            if (javaClass != field.getDeclaringClass()) continue;
-            if (ConstantField.isConstant(field)) constants.add(new ConstantField(field));
-            
-            int modifiers = field.getModifiers();
-            if (Modifier.isStatic(modifiers)) addField(staticCallbacks, staticNames, field, Modifier.isFinal(modifiers), true);
+    private interface Initializer {
+        public void initialize();
+    }
+
+    private class InterfaceInitializer implements Initializer {
+        private volatile boolean hasRun = false;
+        private final Class javaClass;
+
+        public InterfaceInitializer(Class<?> javaClass) {
+            this.javaClass = javaClass;
         }
+        
+        public synchronized void initialize() {
+            if (hasRun) return;
+            hasRun = true;
 
-        // Now add all aliases for the static methods (fields) as appropriate
-        for (Map.Entry<String, NamedInstaller> entry : staticCallbacks.entrySet()) {
-            if (entry.getValue().type == NamedInstaller.STATIC_METHOD && entry.getValue().hasLocalMethod()) {
-                assignAliases((MethodInstaller)entry.getValue(), staticNames);
+            Map<String, AssignedName> staticNames  = new HashMap<String, AssignedName>(STATIC_RESERVED_NAMES);
+            List<ConstantField> constants = new ArrayList<ConstantField>();
+            Map<String, NamedInstaller> staticCallbacks = new HashMap<String, NamedInstaller>();
+            Field[] fields = getDeclaredFields(javaClass);
+
+            for (int i = fields.length; --i >= 0; ) {
+                Field field = fields[i];
+                if (javaClass != field.getDeclaringClass()) continue;
+                if (ConstantField.isConstant(field)) constants.add(new ConstantField(field));
+
+                int modifiers = field.getModifiers();
+                if (Modifier.isStatic(modifiers)) addField(staticCallbacks, staticNames, field, Modifier.isFinal(modifiers), true);
             }
+
+            // Now add all aliases for the static methods (fields) as appropriate
+            for (Map.Entry<String, NamedInstaller> entry : staticCallbacks.entrySet()) {
+                if (entry.getValue().type == NamedInstaller.STATIC_METHOD && entry.getValue().hasLocalMethod()) {
+                    assignAliases((MethodInstaller)entry.getValue(), staticNames);
+                }
+            }
+
+            JavaClass.this.staticAssignedNames = staticNames;
+            JavaClass.this.staticInstallers = staticCallbacks;
+            JavaClass.this.constantFields = constants;
+        }
+    };
+
+    private class ClassInitializer implements Initializer {
+        private volatile boolean hasRun = false;
+        private final Class javaClass;
+
+        public ClassInitializer(Class<?> javaClass) {
+            this.javaClass = javaClass;
         }
 
-        this.staticAssignedNames = staticNames;
-        this.staticInstallers = staticCallbacks;        
-        this.constantFields = constants;
+        public synchronized void initialize() {
+            if (hasRun) return;
+            hasRun = true;
+            
+            Class<?> superclass = javaClass.getSuperclass();
+
+            InitializerState state = new InitializerState(getRuntime(), superclass);
+
+            setupClassFields(javaClass, state);
+            setupClassMethods(javaClass, state);
+            setupClassConstructors(javaClass);
+
+            JavaClass.this.staticAssignedNames = Collections.unmodifiableMap(state.staticNames);
+            JavaClass.this.instanceAssignedNames = Collections.unmodifiableMap(state.instanceNames);
+            JavaClass.this.staticInstallers = Collections.unmodifiableMap(state.staticCallbacks);
+            JavaClass.this.instanceInstallers = Collections.unmodifiableMap(state.instanceCallbacks);
+            JavaClass.this.constantFields = Collections.unmodifiableList(state.constantFields);
+        }
     }
 
-    private void initializeClass(Class<?> javaClass) {
-        Class<?> superclass = javaClass.getSuperclass();
-        Map<String, AssignedName> staticNames;
-        Map<String, AssignedName> instanceNames;
-        if (superclass == null) {
-            staticNames = new HashMap<String, AssignedName>();
-            instanceNames = new HashMap<String, AssignedName>();
-        } else {
-            JavaClass superJavaClass = get(getRuntime(),superclass);
-            staticNames = new HashMap<String, AssignedName>(superJavaClass.getStaticAssignedNames());
-            instanceNames = new HashMap<String, AssignedName>(superJavaClass.getInstanceAssignedNames());
+    private static class InitializerState {
+        public final Map<String, AssignedName> staticNames;
+        public final Map<String, AssignedName> instanceNames;
+        public final Map<String, NamedInstaller> staticCallbacks = new HashMap<String, NamedInstaller>();
+        public final Map<String, NamedInstaller> instanceCallbacks = new HashMap<String, NamedInstaller>();
+        public final List<ConstantField> constantFields = new ArrayList<ConstantField>();
+
+        public InitializerState(Ruby runtime, Class superclass) {
+            if (superclass == null) {
+                staticNames = new HashMap<String, AssignedName>();
+                instanceNames = new HashMap<String, AssignedName>();
+            } else {
+                JavaClass superJavaClass = get(runtime,superclass);
+                staticNames = new HashMap<String, AssignedName>(superJavaClass.getStaticAssignedNames());
+                instanceNames = new HashMap<String, AssignedName>(superJavaClass.getInstanceAssignedNames());
+            }
+            staticNames.putAll(STATIC_RESERVED_NAMES);
+            instanceNames.putAll(INSTANCE_RESERVED_NAMES);
         }
-        staticNames.putAll(STATIC_RESERVED_NAMES);
-        instanceNames.putAll(INSTANCE_RESERVED_NAMES);
-        Map<String, NamedInstaller> staticCallbacks = new HashMap<String, NamedInstaller>();
-        Map<String, NamedInstaller> instanceCallbacks = new HashMap<String, NamedInstaller>();
-        List<ConstantField> constantFields = new ArrayList<ConstantField>();
-        
-        setupClassFields(javaClass, constantFields, staticNames, staticCallbacks, instanceNames, instanceCallbacks);
-        setupClassMethods(javaClass, staticNames, staticCallbacks, instanceNames, instanceCallbacks);
-        setupClassConstructors(javaClass);
-        
-        this.staticAssignedNames = Collections.unmodifiableMap(staticNames);
-        this.instanceAssignedNames = Collections.unmodifiableMap(instanceNames);
-        this.staticInstallers = Collections.unmodifiableMap(staticCallbacks);
-        this.instanceInstallers = Collections.unmodifiableMap(instanceCallbacks);
-        this.constantFields = Collections.unmodifiableList(constantFields);
     }
+
+    private static final Initializer DUMMY_INITIALIZER = new Initializer() {
+        public synchronized void initialize() {
+            // anything useful we could do here?
+        }
+    };
     
     public void setupProxy(final RubyClass proxy) {
+        initializer.initialize();
+
         assert proxyLock.isHeldByCurrentThread();
         proxy.defineFastMethod("__jsend!", __jsend_method);
         final Class<?> javaClass = javaClass();
@@ -524,6 +597,9 @@ public class JavaClass extends JavaObject {
             setupInterfaceProxy(proxy);
             return;
         }
+
+        proxy.setReifiedClass(javaClass);
+        
         assert this.proxyClass == null;
         this.unfinishedProxyClass = proxy;
         if (javaClass.isArray() || javaClass.isPrimitive()) {
@@ -536,7 +612,7 @@ public class JavaClass extends JavaObject {
         installClassFields(proxy);
         installClassMethods(proxy);
         installClassConstructors(proxy);
-        installClassConstants(javaClass, proxy);
+        installClassClasses(javaClass, proxy);
         
         // FIXME: bit of a kludge here (non-interface classes assigned to both
         // class and module fields). simplifies proxy extender code, will go away
@@ -614,20 +690,37 @@ public class JavaClass extends JavaObject {
         }
     }
 
-    private void installClassConstants(final Class<?> javaClass, final RubyClass proxy) {
+    private void installClassClasses(final Class<?> javaClass, final RubyModule proxy) {
         // setup constants for public inner classes
-        Class<?>[] classes = getClasses(javaClass);
+        Class<?>[] classes = getDeclaredClasses(javaClass);
 
-        for (int i = classes.length; --i >= 0;) {
+        for (int i = classes.length; --i >= 0; ) {
             if (javaClass == classes[i].getDeclaringClass()) {
                 Class<?> clazz = classes[i];
+
+                // no non-public inner classes
+                if (!Modifier.isPublic(clazz.getModifiers())) continue;
+                
                 String simpleName = getSimpleName(clazz);
-                if (simpleName.length() == 0) {
-                    continue;
-                }
-                // Ignore bad constant named inner classes pending JRUBY-697
-                if (IdUtil.isConstant(simpleName) && proxy.getConstantAt(simpleName) == null) {
-                    proxy.setConstant(simpleName, Java.get_proxy_class(JAVA_UTILITIES, get(getRuntime(), clazz)));
+                if (simpleName.length() == 0) continue;
+
+                final IRubyObject innerProxy = Java.get_proxy_class(JAVA_UTILITIES,get(getRuntime(),clazz));
+
+                if (IdUtil.isConstant(simpleName)) {
+                    if (proxy.getConstantAt(simpleName) == null) {
+                        proxy.const_set(getRuntime().newString(simpleName), innerProxy);
+                    }
+                } else {
+                    // lower-case name
+                    if (!proxy.respondsTo(simpleName)) {
+                        // define a class method
+                        proxy.getSingletonClass().addMethod(simpleName, new JavaMethodZero(proxy.getSingletonClass(), Visibility.PUBLIC) {
+                            @Override
+                            public IRubyObject call(ThreadContext context, IRubyObject self, RubyModule clazz, String name) {
+                                return innerProxy;
+                            }
+                        });
+                    }
                 }
             }
         }
@@ -649,7 +742,7 @@ public class JavaClass extends JavaObject {
 
     private synchronized void installClassMethods(final RubyClass proxy) {
         for (NamedInstaller installer : staticInstallers.values()) {
-            installer.install(proxy);            
+            installer.install(proxy);
         }
         staticInstallers = null;
         
@@ -692,7 +785,7 @@ public class JavaClass extends JavaObject {
         }
     }
     
-    private void setupClassFields(Class<?> javaClass, List<ConstantField> constantFields, Map<String, AssignedName> staticNames, Map<String, NamedInstaller> staticCallbacks, Map<String, AssignedName> instanceNames, Map<String, NamedInstaller> instanceCallbacks) {
+    private void setupClassFields(Class<?> javaClass, InitializerState state) {
         Field[] fields = getFields(javaClass);
         
         for (int i = fields.length; --i >= 0;) {
@@ -700,20 +793,20 @@ public class JavaClass extends JavaObject {
             if (javaClass != field.getDeclaringClass()) continue;
 
             if (ConstantField.isConstant(field)) {
-                constantFields.add(new ConstantField(field));
+                state.constantFields.add(new ConstantField(field));
                 continue;
             }
 
             int modifiers = field.getModifiers();
             if (Modifier.isStatic(modifiers)) {
-                addField(staticCallbacks, staticNames, field, Modifier.isFinal(modifiers), true);
+                addField(state.staticCallbacks, state.staticNames, field, Modifier.isFinal(modifiers), true);
             } else {
-                addField(instanceCallbacks, instanceNames, field, Modifier.isFinal(modifiers), false);
+                addField(state.instanceCallbacks, state.instanceNames, field, Modifier.isFinal(modifiers), false);
             }
         }
     }
 
-    private void setupClassMethods(Class<?> javaClass, Map<String, AssignedName> staticNames, Map<String, NamedInstaller> staticCallbacks, Map<String, AssignedName> instanceNames, Map<String, NamedInstaller> instanceCallbacks) {
+    private void setupClassMethods(Class<?> javaClass, InitializerState state) {
         // TODO: protected methods.  this is going to require a rework of some of the mechanism.
         Method[] methods = getMethods(javaClass);
 
@@ -722,50 +815,65 @@ public class JavaClass extends JavaObject {
             // install the ones that are named in this class
             Method method = methods[i];
             String name = method.getName();
-            if (Modifier.isStatic(method.getModifiers())) {
-                // Install direct java methods with mangled name so 'send' can call them directly.
-                installStaticMethods(staticCallbacks, javaClass, method, name + METHOD_MANGLE);
-                
-                AssignedName assignedName = staticNames.get(name);
-                if (assignedName == null) {
-                    staticNames.put(name, new AssignedName(name, Priority.METHOD));
-                } else {
-                    if (Priority.METHOD.lessImportantThan(assignedName)) continue;
-                    if (!Priority.METHOD.asImportantAs(assignedName)) {
-                        staticCallbacks.remove(name);
-                        staticCallbacks.remove(name + '=');
-                        staticNames.put(name, new AssignedName(name, Priority.METHOD));
-                    }
-                }
-                installStaticMethods(staticCallbacks, javaClass, method, name);
-            } else {
-                // Install direct java methods with mangled name so 'send' can call them directly.
-                installInstanceMethods(instanceCallbacks, javaClass, method, name + METHOD_MANGLE);
 
-                AssignedName assignedName = instanceNames.get(name);
+            if (Modifier.isStatic(method.getModifiers())) {
+                AssignedName assignedName = state.staticNames.get(name);
+
+                // For JRUBY-4505, restore __method methods for reserved names
+                if (STATIC_RESERVED_NAMES.containsKey(method.getName())) {
+                    installStaticMethods(state.staticCallbacks, javaClass, method, name + METHOD_MANGLE);
+                    continue;
+                }
+
                 if (assignedName == null) {
-                    instanceNames.put(name, new AssignedName(name, Priority.METHOD));
+                    state.staticNames.put(name, new AssignedName(name, Priority.METHOD));
                 } else {
                     if (Priority.METHOD.lessImportantThan(assignedName)) continue;
                     if (!Priority.METHOD.asImportantAs(assignedName)) {
-                        instanceCallbacks.remove(name);
-                        instanceCallbacks.remove(name + '=');
-                        instanceNames.put(name, new AssignedName(name, Priority.METHOD));
+                        state.staticCallbacks.remove(name);
+                        state.staticCallbacks.remove(name + '=');
+                        state.staticNames.put(name, new AssignedName(name, Priority.METHOD));
                     }
                 }
-                installInstanceMethods(instanceCallbacks, javaClass, method, name);
+                installStaticMethods(state.staticCallbacks, javaClass, method, name);
+            } else {
+                AssignedName assignedName = state.instanceNames.get(name);
+
+                // For JRUBY-4505, restore __method methods for reserved names
+                if (INSTANCE_RESERVED_NAMES.containsKey(method.getName())) {
+                    installInstanceMethods(state.instanceCallbacks, javaClass, method, name + METHOD_MANGLE);
+                    continue;
+                }
+
+                if (assignedName == null) {
+                    state.instanceNames.put(name, new AssignedName(name, Priority.METHOD));
+                } else {
+                    if (Priority.METHOD.lessImportantThan(assignedName)) continue;
+                    if (!Priority.METHOD.asImportantAs(assignedName)) {
+                        state.instanceCallbacks.remove(name);
+                        state.instanceCallbacks.remove(name + '=');
+                        state.instanceNames.put(name, new AssignedName(name, Priority.METHOD));
+                    }
+                }
+                installInstanceMethods(state.instanceCallbacks, javaClass, method, name);
             }
         }
 
         // now iterate over all installers and make sure they also have appropriate aliases
-        for (Map.Entry<String, NamedInstaller> entry : staticCallbacks.entrySet()) {
+        for (Map.Entry<String, NamedInstaller> entry : state.staticCallbacks.entrySet()) {
+            // no aliases for __method methods
+            if (entry.getKey().endsWith("__method")) continue;
+
             if (entry.getValue().type == NamedInstaller.STATIC_METHOD && entry.getValue().hasLocalMethod()) {
-                assignAliases((MethodInstaller) entry.getValue(), staticNames);
+                assignAliases((MethodInstaller) entry.getValue(), state.staticNames);
             }
         }
-        for (Map.Entry<String, NamedInstaller> entry : instanceCallbacks.entrySet()) {
+        for (Map.Entry<String, NamedInstaller> entry : state.instanceCallbacks.entrySet()) {
+            // no aliases for __method methods
+            if (entry.getKey().endsWith("__method")) continue;
+            
             if (entry.getValue().type == NamedInstaller.INSTANCE_METHOD && entry.getValue().hasLocalMethod()) {
-                assignAliases((MethodInstaller) entry.getValue(), instanceNames);
+                assignAliases((MethodInstaller) entry.getValue(), state.instanceNames);
             }
         }
     }
@@ -790,6 +898,8 @@ public class JavaClass extends JavaObject {
     
     // old (quasi-deprecated) interface class
     private void setupInterfaceProxy(final RubyClass proxy) {
+        initializer.initialize();
+        
         assert javaClass().isInterface();
         assert proxyLock.isHeldByCurrentThread();
         assert this.proxyClass == null;
@@ -799,6 +909,8 @@ public class JavaClass extends JavaObject {
     }
     
     public void setupInterfaceModule(final RubyModule module) {
+        initializer.initialize();
+        
         assert javaClass().isInterface();
         assert proxyLock.isHeldByCurrentThread();
         assert this.proxyModule == null;
@@ -810,22 +922,8 @@ public class JavaClass extends JavaObject {
         for (NamedInstaller installer : staticInstallers.values()) {
             installer.install(module);
         }
-        // setup constants for public inner classes
-        Class<?>[] classes = getClasses(javaClass);
 
-        for (int i = classes.length; --i >= 0; ) {
-            if (javaClass == classes[i].getDeclaringClass()) {
-                Class<?> clazz = classes[i];
-                String simpleName = getSimpleName(clazz);
-                if (simpleName.length() == 0) continue;
-                
-                // Ignore bad constant named inner classes pending JRUBY-697
-                if (IdUtil.isConstant(simpleName) && module.getConstantAt(simpleName) == null) {
-                    module.const_set(getRuntime().newString(simpleName),
-                        Java.get_proxy_class(JAVA_UTILITIES,get(getRuntime(),clazz)));
-                }
-            }
-        }
+        installClassClasses(javaClass, module);
         
         this.proxyModule = module;
         applyProxyExtenders();
@@ -1049,6 +1147,11 @@ public class JavaClass extends JavaObject {
         return getRuntime().newString(javaClass().getName());
     }
 
+    @JRubyMethod
+    public RubyString inspect() {
+        return getRuntime().newString("class " + javaClass().getName());
+    }
+
     @JRubyMethod
     public IRubyObject canonical_name() {
         String canonicalName = javaClass().getCanonicalName();
@@ -1237,18 +1340,35 @@ public class JavaClass extends JavaObject {
     }
 
     @JRubyMethod(name = "<=>", required = 1)
-    public RubyFixnum op_cmp(IRubyObject other) {
-        if (! (other instanceof JavaClass)) {
-            throw getRuntime().newTypeError("<=> requires JavaClass (" + other.getType() + " given)");
+    public IRubyObject op_cmp(IRubyObject other) {
+        Class me = javaClass();
+        Class them = null;
+
+        // dig out the other class
+        if (other instanceof JavaClass) {
+            JavaClass otherClass = (JavaClass) other;
+            them = otherClass.javaClass();
+        } else if (other instanceof ConcreteJavaProxy) {
+            ConcreteJavaProxy proxy = (ConcreteJavaProxy)other;
+            if (proxy.getObject() instanceof Class) {
+                them = (Class)proxy.getObject();
+            }
         }
-        JavaClass otherClass = (JavaClass) other;
-        if (this.javaClass() == otherClass.javaClass()) {
-            return getRuntime().newFixnum(0);
+
+        if (them != null) {
+            if (this.javaClass() == them) {
+                return getRuntime().newFixnum(0);
+            }
+            if (them.isAssignableFrom(me)) {
+                return getRuntime().newFixnum(-1);
+            }
+            if (me.isAssignableFrom(them)) {
+                return getRuntime().newFixnum(1);
+            }
         }
-        if (otherClass.javaClass().isAssignableFrom(this.javaClass())) {
-            return getRuntime().newFixnum(-1);
-        }
-        return getRuntime().newFixnum(1);
+
+        // can't do a comparison
+        return getRuntime().getNil();
     }
 
     @JRubyMethod
@@ -1283,27 +1403,43 @@ public class JavaClass extends JavaObject {
     }
 
     @JRubyMethod(required = 1, rest = true)
-    public JavaMethod java_method(IRubyObject[] args) throws ClassNotFoundException {
+    public JavaMethod java_method(IRubyObject[] args) {
         String methodName = args[0].asJavaString();
-        Class<?>[] argumentTypes = buildArgumentTypes(args);
-        return JavaMethod.create(getRuntime(), javaClass(), methodName, argumentTypes);
+        try {
+            Class<?>[] argumentTypes = buildArgumentTypes(args);
+            return JavaMethod.create(getRuntime(), javaClass(), methodName, argumentTypes);
+        } catch (ClassNotFoundException cnfe) {
+            throw getRuntime().newNameError("undefined method '" + methodName + "' for class '" + javaClass().getName() + "'",
+                methodName);
+        }
+
     }
 
     @JRubyMethod(required = 1, rest = true)
-    public JavaMethod declared_method(IRubyObject[] args) throws ClassNotFoundException {
+    public JavaMethod declared_method(IRubyObject[] args) {
         String methodName = args[0].asJavaString();
-        Class<?>[] argumentTypes = buildArgumentTypes(args);
-        return JavaMethod.createDeclared(getRuntime(), javaClass(), methodName, argumentTypes);
+        try {
+            Class<?>[] argumentTypes = buildArgumentTypes(args);
+            return JavaMethod.createDeclared(getRuntime(), javaClass(), methodName, argumentTypes);
+        } catch (ClassNotFoundException cnfe) {
+            throw getRuntime().newNameError("undefined method '" + methodName + "' for class '" + javaClass().getName() + "'",
+                methodName);
+        }
     }
 
     @JRubyMethod(required = 1, rest = true)
-    public JavaCallable declared_method_smart(IRubyObject[] args) throws ClassNotFoundException {
+    public JavaCallable declared_method_smart(IRubyObject[] args) {
         String methodName = args[0].asJavaString();
-        Class<?>[] argumentTypes = buildArgumentTypes(args);
- 
-        JavaCallable callable = getMatchingCallable(getRuntime(), javaClass(), methodName, argumentTypes);
 
-        if (callable != null) return callable;
+        try {
+            Class<?>[] argumentTypes = buildArgumentTypes(args);
+
+            JavaCallable callable = getMatchingCallable(getRuntime(), javaClass(), methodName, argumentTypes);
+
+            if (callable != null) return callable;
+        } catch (ClassNotFoundException cnfe) {
+            // fall through to error below
+        }
 
         throw getRuntime().newNameError("undefined method '" + methodName + "' for class '" + javaClass().getName() + "'",
                 methodName);
@@ -1656,7 +1792,7 @@ public class JavaClass extends JavaObject {
         return assignable(javaClass(), otherClass) ? getRuntime().getTrue() : getRuntime().getFalse();
     }
 
-    static boolean assignable(Class<?> thisClass, Class<?> otherClass) {
+    public static boolean assignable(Class<?> thisClass, Class<?> otherClass) {
         if(!thisClass.isPrimitive() && otherClass == Void.TYPE ||
             thisClass.isAssignableFrom(otherClass)) {
             return true;
@@ -1703,6 +1839,14 @@ public class JavaClass extends JavaObject {
             return new Constructor[] {};
         }        
     }
+
+    private static Class<?>[] getDeclaredClasses(Class<?> javaClass) {
+        try {
+            return javaClass.getDeclaredClasses();
+        } catch (SecurityException e) {
+            return new Class<?>[] {};
+        }
+    }
     
     private static Class<?>[] getClasses(Class<?> javaClass) {
         try {
@@ -1728,30 +1872,94 @@ public class JavaClass extends JavaObject {
         }
     }
     
-    private static Method[] getMethods(Class<?> javaClass) {
-        HashMap<String, Method> nameMethods = new HashMap<String, Method>();
+    private static boolean methodsAreEquivalent(Method child, Method parent) {
+        return parent.getDeclaringClass().isAssignableFrom(child.getDeclaringClass())
+                && Arrays.equals(child.getParameterTypes(), parent.getParameterTypes())
+                && child.getReturnType() == parent.getReturnType()
+                && child.isVarArgs() == parent.isVarArgs()
+                && Modifier.isPublic(child.getModifiers()) == Modifier.isPublic(parent.getModifiers())
+                && Modifier.isProtected(child.getModifiers()) == Modifier.isProtected(parent.getModifiers())
+                && Modifier.isStatic(child.getModifiers()) == Modifier.isStatic(parent.getModifiers());
+    }
 
-        // we all all superclasses, but avoid adding superclass methods with same name+signature as subclass methods
-        // see JRUBY-3130
-        for (Class c = javaClass; c != null; c = c.getSuperclass()) {
-            try {
-                for (Method m : c.getDeclaredMethods()) {
-                    int modifiers = m.getModifiers();
-                    if (Modifier.isPublic(modifiers) || Modifier.isProtected(modifiers)) {
-                        String namePlusSig = m.getName() + CodegenUtils.sigParams(m.getParameterTypes());
-                        if (nameMethods.containsKey(namePlusSig)) continue;
-                        nameMethods.put(namePlusSig, m);
+
+    private static void addNewMethods(HashMap<String, List<Method>> nameMethods, Method[] methods, boolean includeStatic, boolean removeDuplicate) {
+        Methods: for (Method m : methods) {
+            if (Modifier.isStatic(m.getModifiers()) && !includeStatic) {
+                // Skip static methods if we're not suppose to include them.
+                // Generally for superclasses; we only bind statics from the actual
+                // class.
+                continue;
+            }
+            List<Method> childMethods = nameMethods.get(m.getName());
+            if (childMethods == null) {
+                // first method of this name, add a collection for it
+                childMethods = new ArrayList<Method>();
+                childMethods.add(m);
+                nameMethods.put(m.getName(), childMethods);
+            } else {
+                // we have seen other methods; check if we already have
+                // an equivalent one
+                for (Method m2 : childMethods) {
+                    if (methodsAreEquivalent(m2, m)) {
+                        if (removeDuplicate) {
+                            // Replace the existing method, since the super call is more general
+                            // and virtual dispatch will call the subclass impl anyway.
+                            // Used for instance methods, for which we usually want to use the highest-up
+                            // callable implementation.
+                            childMethods.remove(m2);
+                            childMethods.add(m);
+                        } else {
+                            // just skip the new method, since we don't need it (already found one)
+                            // used for interface methods, which we want to add unconditionally
+                            // but only if we need them
+                        }
+                        continue Methods;
                     }
                 }
-            } catch (SecurityException e) {
+                // no equivalent; add it
+                childMethods.add(m);
             }
         }
+    }
+    
+    private static Method[] getMethods(Class<?> javaClass) {
+        HashMap<String, List<Method>> nameMethods = new HashMap<String, List<Method>>();
 
-        ArrayList<Method> list2 = new ArrayList<Method>();
-        for (Map.Entry<String, Method> entry : nameMethods.entrySet()) {
-            list2.add(entry.getValue());
+        // we scan all superclasses, but avoid adding superclass methods with
+        // same name+signature as subclass methods (see JRUBY-3130)
+        for (Class c = javaClass; c != null; c = c.getSuperclass()) {
+            // only add class's methods if it's public or we can set accessible
+            // (see JRUBY-4799)
+            if (Modifier.isPublic(c.getModifiers()) || CAN_SET_ACCESSIBLE) {
+                // for each class, scan declared methods for new signatures
+                try {
+                    // add methods, including static if this is the actual class,
+                    // and replacing child methods with equivalent parent methods
+                    addNewMethods(nameMethods, c.getDeclaredMethods(), c == javaClass, true);
+                } catch (SecurityException e) {
+                }
+            }
+
+            // then do the same for each interface
+            for (Class i : c.getInterfaces()) {
+                try {
+                    // add methods, not including static (should be none on
+                    // interfaces anyway) and not replacing child methods with
+                    // parent methods
+                    addNewMethods(nameMethods, i.getMethods(), false, false);
+                } catch (SecurityException e) {
+                }
+            }
         }
         
-        return list2.toArray(new Method[list2.size()]);
+        // now only bind the ones that remain
+        ArrayList<Method> finalList = new ArrayList<Method>();
+
+        for (Map.Entry<String, List<Method>> entry : nameMethods.entrySet()) {
+            finalList.addAll(entry.getValue());
+        }
+        
+        return finalList.toArray(new Method[finalList.size()]);
     }
 }
