@@ -47,13 +47,12 @@ import org.jruby.RubyObject;
 import org.jruby.RubyString;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.anno.JRubyClass;
+import org.jruby.java.proxies.JavaProxy;
 import org.jruby.runtime.Block;
 import org.jruby.runtime.ObjectAllocator;
-import org.jruby.runtime.ObjectMarshal;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.ByteList;
-import org.jruby.util.IOInputStream;
 
 /**
  *
@@ -63,12 +62,24 @@ import org.jruby.util.IOInputStream;
 public class JavaObject extends RubyObject {
 
     private static Object NULL_LOCK = new Object();
+    private final RubyClass.VariableAccessor objectAccessor;
 
     protected JavaObject(Ruby runtime, RubyClass rubyClass, Object value) {
         super(runtime, rubyClass);
+        objectAccessor = rubyClass.getVariableAccessorForWrite("__wrap_struct__");
         dataWrapStruct(value);
     }
 
+    @Override
+    public Object dataGetStruct() {
+        return objectAccessor.get(this);
+    }
+
+    @Override
+    public void dataWrapStruct(Object object) {
+        objectAccessor.set(this, object);
+    }
+
     protected JavaObject(Ruby runtime, Object value) {
         this(runtime, runtime.getJavaSupport().getJavaObjectClass(), value);
     }
@@ -84,6 +95,17 @@ public class JavaObject extends RubyObject {
         return new JavaObject(runtime, value);
     }
 
+    @JRubyMethod(meta = true)
+    public static IRubyObject wrap(ThreadContext context, IRubyObject self, IRubyObject object) {
+        Ruby runtime = context.getRuntime();
+        Object obj = getWrappedObject(object, NEVER);
+
+        if (obj == NEVER) return runtime.getNil();
+
+        return wrap(runtime, obj);
+    }
+
+    @Override
     public Class<?> getJavaClass() {
         Object dataStruct = dataGetStruct();
         return dataStruct != null ? dataStruct.getClass() : Void.TYPE;
@@ -110,11 +132,23 @@ public class JavaObject extends RubyObject {
         result.defineAnnotatedMethods(JavaObject.class);
     }
 
+    @Override
     public boolean equals(Object other) {
-        return other instanceof JavaObject &&
-                this.dataGetStruct() == ((JavaObject) other).dataGetStruct();
+        Ruby runtime = getRuntime();
+        Object myValue = getValue();
+        Object otherValue = other;
+        if (other instanceof IRubyObject) {
+            otherValue = getWrappedObject((IRubyObject)other, NEVER);
+        }
+
+        if (otherValue == NEVER) {
+            // not a wrapped object
+            return false;
+        }
+        return myValue == otherValue;
     }
 
+    @Override
     public int hashCode() {
         Object dataStruct = dataGetStruct();
         if (dataStruct != null) {
@@ -124,51 +158,68 @@ public class JavaObject extends RubyObject {
     }
 
     @JRubyMethod
+    @Override
     public RubyFixnum hash() {
         return getRuntime().newFixnum(hashCode());
     }
 
     @JRubyMethod
+    @Override
     public IRubyObject to_s() {
-        Object dataStruct = dataGetStruct();
+        return to_s(getRuntime(), dataGetStruct());
+    }
+
+    public static IRubyObject to_s(Ruby runtime, Object dataStruct) {
         if (dataStruct != null) {
             String stringValue = dataStruct.toString();
             if (stringValue != null) {
-                return RubyString.newUnicodeString(getRuntime(), dataStruct.toString());
+                return RubyString.newUnicodeString(runtime, dataStruct.toString());
             }
 
-            return getRuntime().getNil();
+            return runtime.getNil();
         }
-        return RubyString.newEmptyString(getRuntime());
+        return RubyString.newEmptyString(runtime);
     }
 
     @JRubyMethod(name = {"==", "eql?"}, required = 1)
     public IRubyObject op_equal(IRubyObject other) {
-        if (!(other instanceof JavaObject)) {
-            other = (JavaObject)other.dataGetStruct();
-            if (!(other instanceof JavaObject)) {
-                return getRuntime().getFalse();
-            }
+        Object myValue = getValue();
+        return opEqualShared(myValue, other);
+    }
+
+    public static IRubyObject op_equal(JavaProxy self, IRubyObject other) {
+        Object myValue = self.getObject();
+        return opEqualShared(myValue, other);
+    }
+
+    private static IRubyObject opEqualShared(Object myValue, IRubyObject other) {
+        Ruby runtime = other.getRuntime();
+        Object otherValue = getWrappedObject(other, NEVER);
+
+        if (other == NEVER) {
+            // not a wrapped object
+            return runtime.getFalse();
         }
 
-        if (getValue() == null && ((JavaObject) other).getValue() == null) {
-            return getRuntime().getTrue();
+        if (myValue == null && otherValue == null) {
+            return runtime.getTrue();
         }
 
-        boolean isEqual = getValue().equals(((JavaObject) other).getValue());
-        return isEqual ? getRuntime().getTrue() : getRuntime().getFalse();
+        return runtime.newBoolean(myValue.equals(otherValue));
     }
 
     @JRubyMethod(name = "equal?", required = 1)
     public IRubyObject same(IRubyObject other) {
-        if (!(other instanceof JavaObject)) {
-            other = (JavaObject)other.dataGetStruct();
-            if (!(other instanceof JavaObject)) {
-                return getRuntime().getFalse();
-            }
+        Ruby runtime = getRuntime();
+        Object myValue = getValue();
+        Object otherValue = getWrappedObject(other, NEVER);
+
+        if (other == NEVER) {
+            // not a wrapped object
+            return runtime.getFalse();
         }
 
-        if (getValue() == null && ((JavaObject) other).getValue() == null) {
+        if (myValue == null && otherValue == null) {
             return getRuntime().getTrue();
         }
 
@@ -176,6 +227,16 @@ public class JavaObject extends RubyObject {
         return isSame ? getRuntime().getTrue() : getRuntime().getFalse();
     }
 
+    private static Object getWrappedObject(IRubyObject other, Object def) {
+        if (other instanceof JavaObject) {
+            return ((JavaObject)other).getValue();
+        } else if (other instanceof JavaProxy) {
+            return ((JavaProxy)other).getObject();
+        } else {
+            return def;
+        }
+    }
+
     @JRubyMethod
     public RubyString java_type() {
         return getRuntime().newString(getJavaClass().getName());
@@ -219,6 +280,12 @@ public class JavaObject extends RubyObject {
         }
     }
     
+    public static IRubyObject ruby_synchronized(ThreadContext context, Object lock, Block block) {
+        synchronized (lock != null ? lock : NULL_LOCK) {
+            return block.yield(context, null);
+        }
+    }
+    
     @JRubyMethod(frame = true)
     public IRubyObject marshal_dump() {
         if (Serializable.class.isAssignableFrom(getJavaClass())) {
@@ -241,7 +308,7 @@ public class JavaObject extends RubyObject {
     public IRubyObject marshal_load(ThreadContext context, IRubyObject str) {
         try {
             ByteList byteList = str.convertToString().getByteList();
-            ByteArrayInputStream bais = new ByteArrayInputStream(byteList.bytes, byteList.begin, byteList.realSize);
+            ByteArrayInputStream bais = new ByteArrayInputStream(byteList.getUnsafeBytes(), byteList.getBegin(), byteList.getRealSize());
             ObjectInputStream ois = new ObjectInputStream(bais);
 
             dataWrapStruct(ois.readObject());
@@ -254,6 +321,19 @@ public class JavaObject extends RubyObject {
         }
     }
 
+    @Override
+    public Object toJava(Class cls) {
+        if (getValue() == null) {
+            // THIS SHOULD NEVER HAPPEN, but it DOES
+            return getValue();
+        }
+        
+        if (cls.isAssignableFrom(getValue().getClass())) {
+            return getValue();
+        }
+        
+        return super.toJava(cls);
+    }
 
     private static final ObjectAllocator JAVA_OBJECT_ALLOCATOR = new ObjectAllocator() {
         public IRubyObject allocate(Ruby runtime, RubyClass klazz) {
