@@ -42,13 +42,12 @@ import org.jruby.runtime.scope.ManyVarsDynamicScope;
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
 import org.jruby.RubyClass;
-import org.jruby.RubyKernel.CatchTarget;
+import org.jruby.RubyContinuation.Continuation;
 import org.jruby.RubyModule;
 import org.jruby.RubyString;
 import org.jruby.RubyThread;
 import org.jruby.evaluator.ASTInterpreter;
 import org.jruby.exceptions.JumpException.ReturnJump;
-import org.jruby.internal.runtime.JumpTarget;
 import org.jruby.internal.runtime.methods.DefaultMethod;
 import org.jruby.lexer.yacc.ISourcePosition;
 import org.jruby.libraries.FiberLibrary.Fiber;
@@ -58,16 +57,16 @@ import org.jruby.parser.StaticScope;
 import org.jruby.runtime.builtin.IRubyObject;
 
 public final class ThreadContext {
-    public static synchronized ThreadContext newContext(Ruby runtime) {
+    public static ThreadContext newContext(Ruby runtime) {
         ThreadContext context = new ThreadContext(runtime);
-        //        if(runtime.getInstanceConfig().isSamplingEnabled()) {
-        //    org.jruby.util.SimpleSampler.registerThreadContext(context);
-        //}
 
         return context;
     }
     
     private final static int INITIAL_SIZE = 10;
+    
+    /** The number of calls after which to do a thread event poll */
+    private final static int CALL_POLL_COUNT = 0xFFF;
     private final static String UNKNOWN_NAME = "(unknown)";
     
     private final Ruby runtime;
@@ -81,11 +80,9 @@ public final class ThreadContext {
     private RubyThread thread;
     private Fiber fiber;
     
-    //private UnsynchronizedStack parentStack;
     private RubyModule[] parentStack = new RubyModule[INITIAL_SIZE];
     private int parentIndex = -1;
     
-    //private UnsynchronizedStack frameStack;
     private Frame[] frameStack = new Frame[INITIAL_SIZE];
     private int frameIndex = -1;
     
@@ -93,8 +90,9 @@ public final class ThreadContext {
     // to implement closures.
     private DynamicScope[] scopeStack = new DynamicScope[INITIAL_SIZE];
     private int scopeIndex = -1;
-    
-    private CatchTarget[] catchStack = new CatchTarget[INITIAL_SIZE];
+
+    private static final Continuation[] EMPTY_CATCHTARGET_STACK = new Continuation[0];
+    private Continuation[] catchStack = EMPTY_CATCHTARGET_STACK;
     private int catchIndex = -1;
     
     // File where current executing unit is being evaluated
@@ -109,6 +107,7 @@ public final class ThreadContext {
     // the grep is running in so that the backref will be set in an
     // appropriate place.
     private int rubyFrameDelta = 0;
+    private boolean eventHooksEnabled = true;
     
     /**
      * Constructor for Context.
@@ -120,9 +119,11 @@ public final class ThreadContext {
         // away and then pass it into top-level parse so it ends up being the top level.
         StaticScope topStaticScope = new LocalStaticScope(null);
         pushScope(new ManyVarsDynamicScope(topStaticScope, null));
-            
-        for (int i = 0; i < frameStack.length; i++) {
-            frameStack[i] = new Frame();
+
+        Frame[] stack = frameStack;
+        int length = stack.length;
+        for (int i = 0; i < length; i++) {
+            stack[i] = new Frame();
         }
     }
 
@@ -254,6 +255,11 @@ public final class ThreadContext {
     
     public void setThread(RubyThread thread) {
         this.thread = thread;
+
+        // associate the thread with this context, unless we're clearing the reference
+        if (thread != null) {
+            thread.setContext(this);
+        }
     }
     
     public Fiber getFiber() {
@@ -264,46 +270,41 @@ public final class ThreadContext {
         this.fiber = fiber;
     }
     
-//    public IRubyObject getLastline() {
-//        IRubyObject value = getCurrentScope().getLastLine();
-//        
-//        // DynamicScope does not preinitialize these values since they are virtually never used.
-//        return value == null ? runtime.getNil() : value;
-//    }
-//    
-//    public void setLastline(IRubyObject value) {
-//        getCurrentScope().setLastLine(value);
-//    }
-    
     //////////////////// CATCH MANAGEMENT ////////////////////////
     private void expandCatchIfNecessary() {
         int newSize = catchStack.length * 2;
-        CatchTarget[] newCatchStack = new CatchTarget[newSize];
+        if (newSize == 0) newSize = 1;
+        Continuation[] newCatchStack = new Continuation[newSize];
 
         System.arraycopy(catchStack, 0, newCatchStack, 0, catchStack.length);
         catchStack = newCatchStack;
     }
     
-    public void pushCatch(CatchTarget catchTarget) {
+    public void pushCatch(Continuation catchTarget) {
         int index = ++catchIndex;
-        CatchTarget[] stack = catchStack;
-        stack[index] = catchTarget;
-        if (index + 1 == stack.length) {
+        if (index == catchStack.length) {
             expandCatchIfNecessary();
         }
+        catchStack[index] = catchTarget;
     }
     
     public void popCatch() {
         catchIndex--;
     }
-    
-    public CatchTarget[] getActiveCatches() {
-        int index = catchIndex;
-        if (index < 0) return new CatchTarget[0];
-        
-        CatchTarget[] activeCatches = new CatchTarget[index + 1];
-        System.arraycopy(catchStack, 0, activeCatches, 0, index + 1);
-        return activeCatches;
+
+    /**
+     * Find the active Continuation for the given tag. Must be called with an
+     * interned string.
+     *
+     * @param tag The interned string to search for
+     * @return The continuation associated with this tag
+     */
+    public Continuation getActiveCatch(String tag) {
+        for (int i = catchIndex; i >= 0; i--) {
+            Continuation c = catchStack[i];
+            if (c.tag == tag) return c;
+        }
+        return null;
     }
     
     //////////////////// FRAME MANAGEMENT ////////////////////////
@@ -331,7 +332,7 @@ public final class ThreadContext {
                                IRubyObject self, Block block) {
         int index = ++this.frameIndex;
         Frame[] stack = frameStack;
-        stack[index].updateFrame(clazz, self, name, block, file, line);
+        stack[index].updateFrame(clazz, self, name, block, file, line, callNumber);
         if (index + 1 == stack.length) {
             expandFramesIfNecessary();
         }
@@ -340,7 +341,7 @@ public final class ThreadContext {
     private void pushEvalFrame(IRubyObject self) {
         int index = ++this.frameIndex;
         Frame[] stack = frameStack;
-        stack[index].updateFrameForEval(self, file, line);
+        stack[index].updateFrameForEval(self, file, line, callNumber);
         if (index + 1 == stack.length) {
             expandFramesIfNecessary();
         }
@@ -425,6 +426,22 @@ public final class ThreadContext {
         }
         return frames;
     }
+
+    /**
+     * Search the frame stack for the given JumpTarget. Return true if it is
+     * found and false otherwise. Skip the given number of frames before
+     * beginning the search.
+     * 
+     * @param target The JumpTarget to search for
+     * @param skipFrames The number of frames to skip before searching
+     * @return
+     */
+    public boolean isJumpTargetAlive(int target, int skipFrames) {
+        for (int i = frameIndex - skipFrames; i >= 0; i--) {
+            if (frameStack[i].getJumpTarget() == target) return true;
+        }
+        return false;
+    }
     
     public String getFrameName() {
         return getCurrentFrame().getName();
@@ -434,14 +451,10 @@ public final class ThreadContext {
         return getCurrentFrame().getSelf();
     }
     
-    public JumpTarget getFrameJumpTarget() {
+    public int getFrameJumpTarget() {
         return getCurrentFrame().getJumpTarget();
     }
     
-    @Deprecated
-    public void setFrameJumpTarget(JumpTarget target) {
-    }
-    
     public RubyModule getFrameKlazz() {
         return getCurrentFrame().getKlazz();
     }
@@ -497,13 +510,25 @@ public final class ThreadContext {
         thread.pollThreadEvents(this);
     }
     
-    int calls = 0;
+    public int callNumber = 0;
+
+    public int getCurrentTarget() {
+        return callNumber;
+    }
     
     public void callThreadPoll() {
-        if ((calls++ & 0xFF) == 0) pollThreadEvents();
+        if ((callNumber++ & CALL_POLL_COUNT) == 0) pollThreadEvents();
+    }
+
+    public static void callThreadPoll(ThreadContext context) {
+        if ((context.callNumber++ & CALL_POLL_COUNT) == 0) context.pollThreadEvents();
     }
     
     public void trace(RubyEvent event, String name, RubyModule implClass) {
+        trace(event, name, implClass, file, line);
+    }
+
+    public void trace(RubyEvent event, String name, RubyModule implClass, String file, int line) {
         runtime.callEventHooks(this, event, file, line, name, implClass);
     }
     
@@ -768,6 +793,14 @@ public final class ThreadContext {
         return traceFrames;
     }
 
+    public boolean isEventHooksEnabled() {
+        return eventHooksEnabled;
+    }
+
+    public void setEventHooksEnabled(boolean flag) {
+        eventHooksEnabled = flag;
+    }
+
     public static class RubyStackTraceElement {
         private StackTraceElement element;
         private boolean binding;
@@ -1073,6 +1106,7 @@ public final class ThreadContext {
         pushFrameCopy();
         getCurrentFrame().setSelf(type);
         getCurrentFrame().setVisibility(Visibility.PUBLIC);
+
         pushScope(DynamicScope.newDynamicScope(staticScope, null));
     }
     
