@@ -11,6 +11,7 @@
  * implied. See the License for the specific language governing
  * rights and limitations under the License.
  *
+ * Copyright (C) 2002-2010 JRuby Community
  * Copyright (C) 2002-2004 Anders Bengtsson <ndrsbngtssn@yahoo.se>
  * Copyright (C) 2002-2004 Jan Arne Petersen <jpetersen@uni-bonn.de>
  * Copyright (C) 2004 Thomas E Enebo <enebo@acm.org>
@@ -44,12 +45,12 @@ import java.util.Hashtable;
 import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
+import java.util.concurrent.atomic.AtomicInteger;
 import java.util.jar.JarFile;
 import java.util.regex.Matcher;
 import java.util.regex.Pattern;
-
 import java.util.zip.ZipException;
-import org.jruby.CompatVersion;
+
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
 import org.jruby.RubyFile;
@@ -124,6 +125,8 @@ import org.jruby.util.JRubyFile;
  * @author jpetersen
  */
 public class LoadService {
+    private final LoadTimer loadTimer;
+
     public enum SuffixType {
         Source, Extension, Both, Neither;
         
@@ -162,6 +165,11 @@ public class LoadService {
     
     public LoadService(Ruby runtime) {
         this.runtime = runtime;
+        if (RubyInstanceConfig.DEBUG_LOAD_TIMINGS) {
+            loadTimer = new TracingLoadTimer();
+        } else {
+            loadTimer = new LoadTimer();
+        }
     }
 
     public void init(List additionalDirectories) {
@@ -175,39 +183,37 @@ public class LoadService {
         }
 
         // add $RUBYLIB paths
-       RubyHash env = (RubyHash) runtime.getObject().fastGetConstant("ENV");
-       RubyString env_rubylib = runtime.newString("RUBYLIB");
-       if (env.has_key_p(env_rubylib).isTrue()) {
-           String rubylib = env.op_aref(runtime.getCurrentContext(), env_rubylib).toString();
-           String[] paths = rubylib.split(File.pathSeparator);
-           for (int i = 0; i < paths.length; i++) {
-               addPath(paths[i]);
-           }
-       }
+        RubyHash env = (RubyHash) runtime.getObject().fastGetConstant("ENV");
+        RubyString env_rubylib = runtime.newString("RUBYLIB");
+        if (env.has_key_p(env_rubylib).isTrue()) {
+            String rubylib = env.op_aref(runtime.getCurrentContext(), env_rubylib).toString();
+            String[] paths = rubylib.split(File.pathSeparator);
+            for (int i = 0; i < paths.length; i++) {
+                addPath(paths[i]);
+            }
+        }
 
-       // wrap in try/catch for security exceptions in an applet
-       if (!Ruby.isSecurityRestricted()) {
-           try {
-               String jrubyHome = runtime.getJRubyHome();
-               if (jrubyHome != null) {
-                   char sep = '/';
-                   String rubyDir = jrubyHome + sep + "lib" + sep + "ruby" + sep;
+        // wrap in try/catch for security exceptions in an applet
+        try {
+            String jrubyHome = runtime.getJRubyHome();
+            if (jrubyHome != null) {
+                char sep = '/';
+                String rubyDir = jrubyHome + sep + "lib" + sep + "ruby" + sep;
 
-                   // If we're running in 1.9 compat mode, add Ruby 1.9 libs to path before 1.8 libs
-                   if (runtime.is1_9()) {
-                       addPath(rubyDir + "site_ruby" + sep + Constants.RUBY1_9_MAJOR_VERSION);
-                       addPath(rubyDir + "site_ruby" + sep + "shared");
-                       addPath(rubyDir + "site_ruby" + sep + Constants.RUBY_MAJOR_VERSION);
-                       addPath(rubyDir + Constants.RUBY1_9_MAJOR_VERSION);
-                   } else {
-                       // Add 1.8 libs
-                       addPath(rubyDir + "site_ruby" + sep + Constants.RUBY_MAJOR_VERSION);
-                       addPath(rubyDir + "site_ruby" + sep + "shared");
-                       addPath(rubyDir + Constants.RUBY_MAJOR_VERSION);
-                   }
-               }
-           } catch(SecurityException e) {}
-       }
+                // If we're running in 1.9 compat mode, add Ruby 1.9 libs to path before 1.8 libs
+                if (runtime.is1_9()) {
+                    addPath(rubyDir + "site_ruby" + sep + Constants.RUBY1_9_MAJOR_VERSION);
+                    addPath(rubyDir + "site_ruby" + sep + "shared");
+                    addPath(rubyDir + "site_ruby" + sep + Constants.RUBY_MAJOR_VERSION);
+                    addPath(rubyDir + Constants.RUBY1_9_MAJOR_VERSION);
+                } else {
+                    // Add 1.8 libs
+                    addPath(rubyDir + "site_ruby" + sep + Constants.RUBY_MAJOR_VERSION);
+                    addPath(rubyDir + "site_ruby" + sep + "shared");
+                    addPath(rubyDir + Constants.RUBY_MAJOR_VERSION);
+                }
+            }
+        } catch(SecurityException ignore) {}
         
         // "." dir is used for relative path loads from a given file, as in require '../foo/bar'
         if (runtime.getSafeLevel() == 0) {
@@ -249,7 +255,7 @@ public class LoadService {
             library.load(runtime, wrap);
         } catch (IOException e) {
             if (runtime.getDebug().isTrue()) e.printStackTrace(runtime.getErr());
-            throw runtime.newLoadError("IO error -- " + file);
+            throw newLoadErrorFromThrowable(runtime, file, e);
         }
     }
 
@@ -312,11 +318,47 @@ public class LoadService {
         }
     }
 
+    private static class LoadTimer {
+        public long startLoad(String file) { return 0L; }
+        public void endLoad(String file, long startTime) {}
+    }
+
+    private static class TracingLoadTimer extends LoadTimer {
+        private final AtomicInteger indent = new AtomicInteger(0);
+        private String getIndentString() {
+            StringBuilder buf = new StringBuilder();
+            int i = indent.get();
+            for (int j = 0; j < i; j++) {
+                buf.append("  ");
+            }
+            return buf.toString();
+        }
+        @Override
+        public long startLoad(String file) {
+            int i = indent.incrementAndGet();
+            System.err.println(getIndentString() + "-> " + file);
+            return System.currentTimeMillis();
+        }
+        @Override
+        public void endLoad(String file, long startTime) {
+            System.err.println(getIndentString() + "<- " + file + " - "
+                    + (System.currentTimeMillis() - startTime) + "ms");
+            indent.decrementAndGet();
+        }
+    }
+
     public boolean require(String file) {
         if(!runtime.getProfile().allowRequire(file)) {
             throw runtime.newLoadError("No such file to load -- " + file);
         }
-        return smartLoad(file);
+
+        long startTime = loadTimer.startLoad(file);
+        try {
+            return smartLoad(file);
+        } finally {
+            loadTimer.endLoad(file, startTime);
+        }
+
     }
 
     /**
@@ -502,7 +544,7 @@ public class LoadService {
                 state.library = new ClassExtensionLibrary(theClass);
             } catch (Exception ee) {
                 state.library = null;
-                runtime.getGlobalVariables().set("$!", runtime.getNil());
+                runtime.getGlobalVariables().clear("$!");
             }
 
             // If there was a good library before, we go back to that
@@ -551,7 +593,7 @@ public class LoadService {
             state.library = new ScriptClassLibrary(script);
         }
     }
-    
+
     public class SearchState {
         public Library library;
         public String loadName;
@@ -641,11 +683,15 @@ public class LoadService {
 
             if(runtime.getDebug().isTrue()) e.printStackTrace(runtime.getErr());
             
-            RaiseException re = runtime.newLoadError("IO error -- " + state.searchFile);
+            RaiseException re = newLoadErrorFromThrowable(runtime, state.searchFile, e);
             re.initCause(e);
             throw re;
         }
     }
+
+    private static RaiseException newLoadErrorFromThrowable(Ruby runtime, String file, Throwable t) {
+        return runtime.newLoadError(String.format("load error: %s -- %s: %s", file, t.getClass().getName(), t.getMessage()));
+    }
     
     protected final List<LoadSearcher> searchers = new ArrayList<LoadSearcher>();
     {
@@ -745,7 +791,7 @@ public class LoadService {
             String file = baseName + suffix;
             LoadServiceResource resource = findFileInClasspath(file);
             if (resource != null) {
-                state.loadName = file;
+                state.loadName = resolveLoadName(resource, file);
                 return createLibrary(state, resource);
             }
         }
@@ -785,7 +831,7 @@ public class LoadService {
                     }
                     foundResource = new LoadServiceResource(file, s, absolute);
                     debugLogFound(foundResource);
-                    state.loadName = namePlusSuffix;
+                    state.loadName = resolveLoadName(foundResource, namePlusSuffix);
                     break;
                 }
             } catch (IllegalArgumentException illArgEx) {
@@ -795,6 +841,40 @@ public class LoadService {
         
         return foundResource;
     }
+
+    protected LoadServiceResource tryResourceFromHome(SearchState state, String baseName, SuffixType suffixType) throws RaiseException {
+        LoadServiceResource foundResource = null;
+
+        RubyHash env = (RubyHash) runtime.getObject().fastGetConstant("ENV");
+        RubyString env_home = runtime.newString("HOME");
+        if (env.has_key_p(env_home).isFalse()) {
+            return null;
+        }
+        String home = env.op_aref(runtime.getCurrentContext(), env_home).toString();
+        String path = baseName.substring(2);
+
+        for (String suffix : suffixType.getSuffixes()) {
+            String namePlusSuffix = path + suffix;
+            // check home directory; if file exists, retrieve URL and return resource
+            try {
+                JRubyFile file = JRubyFile.create(home, RubyFile.expandUserPath(runtime.getCurrentContext(), namePlusSuffix));
+                debugLogTry("resourceFromHome", file.toString());
+                if (file.isFile() && file.isAbsolute() && file.canRead()) {
+                    boolean absolute = true;
+                    String s = "~/" + namePlusSuffix;
+                    
+                    foundResource = new LoadServiceResource(file, s, absolute);
+                    debugLogFound(foundResource);
+                    state.loadName = resolveLoadName(foundResource, s);
+                    break;
+                }
+            } catch (IllegalArgumentException illArgEx) {
+            } catch (SecurityException secEx) {
+            }
+        }
+
+        return foundResource;
+    }
     
     protected LoadServiceResource tryResourceFromJarURL(SearchState state, String baseName, SuffixType suffixType) {
         // if a jar or file URL, return load service resource directly without further searching
@@ -816,7 +896,7 @@ public class LoadService {
                     throw runtime.newIOErrorFromException(e);
                 }
                 if (foundResource != null) {
-                    state.loadName = namePlusSuffix;
+                    state.loadName = resolveLoadName(foundResource, namePlusSuffix);
                     break; // end suffix iteration
                 }
             }
@@ -836,7 +916,7 @@ public class LoadService {
                     }
                 } catch(Exception e) {}
                 if (foundResource != null) {
-                    state.loadName = namePlusSuffix;
+                    state.loadName = resolveLoadName(foundResource, namePlusSuffix);
                     break; // end suffix iteration
                 }
             }    
@@ -853,7 +933,17 @@ public class LoadService {
             foundResource = tryResourceFromCWD(state, baseName, suffixType);
 
             if (foundResource != null) {
-                state.loadName = foundResource.getName();
+                state.loadName = resolveLoadName(foundResource, foundResource.getName());
+                return foundResource;
+            }
+        }
+
+        // if it's a ~/ baseName use HOME logic
+        if (baseName.startsWith("~/")) {
+            foundResource = tryResourceFromHome(state, baseName, suffixType);
+
+            if (foundResource != null) {
+                state.loadName = resolveLoadName(foundResource, foundResource.getName());
                 return foundResource;
             }
         }
@@ -865,7 +955,7 @@ public class LoadService {
                 foundResource = tryResourceAsIs(namePlusSuffix);
 
                 if (foundResource != null) {
-                    state.loadName = namePlusSuffix;
+                    state.loadName = resolveLoadName(foundResource, namePlusSuffix);
                     return foundResource;
                 }
             }
@@ -873,12 +963,13 @@ public class LoadService {
             return null;
         }
         
-        Outer: for (Iterator pathIter = loadPath.getList().iterator(); pathIter.hasNext();) {
-            // TODO this is really ineffient, and potentially a problem everytime anyone require's something.
+        Outer: for (int i = 0; i < loadPath.size(); i++) {
+            // TODO this is really inefficient, and potentially a problem everytime anyone require's something.
             // we should try to make LoadPath a special array object.
-            String loadPathEntry = ((IRubyObject)pathIter.next()).toString();
+            RubyString entryString = loadPath.eltInternal(i).convertToString();
+            String loadPathEntry = entryString.asJavaString();
 
-            if (loadPathEntry.equals(".")) {
+            if (loadPathEntry.equals(".") || loadPathEntry.equals("")) {
                 foundResource = tryResourceFromCWD(state, baseName, suffixType);
 
                 if (foundResource != null) {
@@ -886,15 +977,18 @@ public class LoadService {
                     if(ss.startsWith("./")) {
                         ss = ss.substring(2);
                     }
-                    state.loadName = ss;
+                    state.loadName = resolveLoadName(foundResource, ss);
                     break Outer;
                 }
             } else {
+                boolean looksLikeJarURL = loadPathLooksLikeJarURL(loadPathEntry);
                 for (String suffix : suffixType.getSuffixes()) {
                     String namePlusSuffix = baseName + suffix;
 
-                    if (loadPathLooksLikeJarURL(loadPathEntry)) {
+                    if (looksLikeJarURL) {
                         foundResource = tryResourceFromJarURLWithLoadPath(namePlusSuffix, loadPathEntry);
+                    } else if(namePlusSuffix.startsWith("./")) {
+                        throw runtime.newLoadError("");
                     } else {
                         foundResource = tryResourceFromLoadPath(namePlusSuffix, loadPathEntry);
                     }
@@ -904,7 +998,7 @@ public class LoadService {
                         if(ss.startsWith("./")) {
                             ss = ss.substring(2);
                         }
-                        state.loadName = ss;
+                        state.loadName = resolveLoadName(foundResource, ss);
                         break Outer; // end suffix iteration
                     }
                 }
@@ -933,7 +1027,7 @@ public class LoadService {
                 }
                 jarFiles.put(loadPathEntry,current);
             } catch (ZipException ignored) {
-                if (runtime.getInstanceConfig().isVerbose()) {
+                if (runtime.getInstanceConfig().isDebug()) {
                     runtime.getErr().println("ZipException trying to access " + loadPathEntry + ", stack trace follows:");
                     ignored.printStackTrace(runtime.getErr());
                 }
@@ -989,7 +1083,7 @@ public class LoadService {
                     }
                     actualPath = JRubyFile.create(JRubyFile.create(runtime.getCurrentDirectory(), loadPathEntry).getAbsolutePath(), RubyFile.expandUserPath(runtime.getCurrentContext(), namePlusSuffix));
                 }
-                debugLogTry("resourceFromLoadPath", actualPath.toString());
+                debugLogTry("resourceFromLoadPath", "'" + actualPath.toString() + "' " + actualPath.isFile() + " " + actualPath.canRead());
                 if (actualPath.isFile() && actualPath.canRead()) {
                     foundResource = new LoadServiceResource(actualPath, reportedPath, absolute);
                     debugLogFound(foundResource);
@@ -1017,11 +1111,8 @@ public class LoadService {
                     if (reportedPath.charAt(0) == '.' && reportedPath.charAt(1) == '/') {
                         reportedPath = reportedPath.replaceFirst("\\./", runtime.getCurrentDirectory());
                     }
-//                     if (reportedPath.charAt(0) != '.') {
-//                         reportedPath = "./" + reportedPath;
-//                     }
+
                     actualPath = JRubyFile.create(runtime.getCurrentDirectory(), RubyFile.expandUserPath(runtime.getCurrentContext(), namePlusSuffix));
-                    //                    actualPath = new File(RubyFile.expandUserPath(runtime.getCurrentContext(), reportedPath));
                 }
                 debugLogTry("resourceAsIs", actualPath.toString());
                 if (actualPath.isFile() && actualPath.canRead()) {
@@ -1053,23 +1144,38 @@ public class LoadService {
             classLoader = runtime.getInstanceConfig().getLoader();
         }
 
-        for (Iterator pathIter = loadPath.getList().iterator(); pathIter.hasNext();) {
-            String entry = pathIter.next().toString();
+        // absolute classpath URI, no need to iterate over loadpaths
+        if (name.startsWith("classpath:/")) {
+            LoadServiceResource foundResource = getClassPathResource(classLoader, name);
+            if (foundResource != null) {
+                return foundResource;
+            }
+        } else if (name.startsWith("classpath:")) {
+            // "relative" classpath URI
+            name = name.substring("classpath:".length());
+        }
+
+        for (int i = 0; i < loadPath.size(); i++) {
+            // TODO this is really inefficient, and potentially a problem everytime anyone require's something.
+            // we should try to make LoadPath a special array object.
+            RubyString entryString = loadPath.eltInternal(i).convertToString();
+            String entry = entryString.asJavaString();
 
             // if entry is an empty string, skip it
             if (entry.length() == 0) continue;
 
             // if entry starts with a slash, skip it since classloader resources never start with a /
             if (entry.charAt(0) == '/' || (entry.length() > 1 && entry.charAt(1) == ':')) continue;
-            
-            // otherwise, try to load from classpath (Note: Jar resources always uses '/')
-            debugLogTry("fileInClasspath", entry + "/" + name);
-            URL loc = classLoader.getResource(entry + "/" + name);
 
-            // Make sure this is not a directory or unavailable in some way
-            if (isRequireable(loc)) {
-                LoadServiceResource foundResource = new LoadServiceResource(loc, loc.getPath());
-                debugLogFound(foundResource);
+            if (entry.startsWith("classpath:/")) {
+                entry = entry.substring("classpath:/".length());
+            } else if (entry.startsWith("classpath:")) {
+                entry = entry.substring("classpath:".length());
+            }
+
+            // otherwise, try to load from classpath (Note: Jar resources always uses '/')
+            LoadServiceResource foundResource = getClassPathResource(classLoader, entry + "/" + name);
+            if (foundResource != null) {
                 return foundResource;
             }
         }
@@ -1079,12 +1185,8 @@ public class LoadService {
         
         // Try to load from classpath without prefix. "A/b.rb" will not load as 
         // "./A/b.rb" in a jar file.
-        debugLogTry("fileInClasspath", name);
-        URL loc = classLoader.getResource(name);
-
-        if (isRequireable(loc)) {
-            LoadServiceResource foundResource = new LoadServiceResource(loc, loc.getPath());
-            debugLogFound(foundResource);
+        LoadServiceResource foundResource = getClassPathResource(classLoader, name);
+        if (foundResource != null) {
             return foundResource;
         }
 
@@ -1105,6 +1207,44 @@ public class LoadService {
         }
         return false;
     }
+
+    protected LoadServiceResource getClassPathResource(ClassLoader classLoader, String name) {
+        boolean isClasspathScheme = false;
+
+        // strip the classpath scheme first
+        if (name.startsWith("classpath:/")) {
+            isClasspathScheme = true;
+            name = name.substring("classpath:/".length());
+        } else if (name.startsWith("classpath:")) {
+            isClasspathScheme = true;
+            name = name.substring("classpath:".length());
+        }
+
+        debugLogTry("fileInClasspath", name);
+        URL loc = classLoader.getResource(name);
+
+        if (loc != null) { // got it
+            String path = "classpath:/" + name;
+            // special case for typical jar:file URLs, but only if the name didn't have
+            // the classpath scheme explicitly
+            if (!isClasspathScheme && loc.toString().startsWith("jar:file:") && isRequireable(loc)) {
+                // Make sure this is not a directory or unavailable in some way
+                try {
+                    path = loc.toURI().getSchemeSpecificPart();
+                } catch (java.net.URISyntaxException urise) {
+                    if (runtime.getInstanceConfig().isDebug()) {
+                        runtime.getErr().println("URISyntaxException trying to parse " + loc + ", stack trace follows:");
+                        urise.printStackTrace(runtime.getErr());
+                    }
+                    return null;
+                }
+            }
+            LoadServiceResource foundResource = new LoadServiceResource(loc, path);
+            debugLogFound(foundResource);
+            return foundResource;
+        }
+        return null;
+    }
     
     protected String canonicalizePath(String path) {
         try {
@@ -1116,4 +1256,8 @@ public class LoadService {
         return path;
       }
     }
+
+    protected String resolveLoadName(LoadServiceResource foundResource, String previousPath) {
+        return previousPath;
+    }
 }
