@@ -96,4 +96,9 @@ public final class KCode {
         }
         return encoding;
     }
+
+    @Override
+    public String toString() {
+        return name();
+    }
 }
