@@ -28,6 +28,11 @@
  ***** END LICENSE BLOCK *****/
 package org.jruby.util.io;
 
+import static java.util.logging.Logger.getLogger;
+import static org.jruby.util.io.ModeFlags.RDONLY;
+import static org.jruby.util.io.ModeFlags.RDWR;
+import static org.jruby.util.io.ModeFlags.WRONLY;
+
 import java.io.File;
 import java.io.FileDescriptor;
 import java.io.FileNotFoundException;
@@ -40,17 +45,16 @@ import java.nio.channels.Channel;
 import java.nio.channels.Channels;
 import java.nio.channels.FileChannel;
 import java.nio.channels.ReadableByteChannel;
-import java.nio.channels.SocketChannel;
 import java.nio.channels.WritableByteChannel;
 import java.util.concurrent.atomic.AtomicInteger;
 import java.util.jar.JarFile;
-import static java.util.logging.Logger.getLogger;
 import java.util.zip.ZipEntry;
+import org.jruby.RubyFile;
+
 import org.jruby.RubyIO;
 import org.jruby.ext.posix.POSIX;
 import org.jruby.util.ByteList;
 import org.jruby.util.JRubyFile;
-import static org.jruby.util.io.ModeFlags.*;
 
 /**
  * ChannelDescriptor provides an abstraction similar to the concept of a
@@ -436,7 +440,7 @@ public class ChannelDescriptor {
         checkOpen();
         
         byteList.ensure(byteList.length() + number);
-        int bytesRead = read(ByteBuffer.wrap(byteList.unsafeBytes(), 
+        int bytesRead = read(ByteBuffer.wrap(byteList.getUnsafeBytes(),
                 byteList.begin() + byteList.length(), number));
         if (bytesRead > 0) {
             byteList.length(byteList.length() + bytesRead);
@@ -525,7 +529,7 @@ public class ChannelDescriptor {
     public int write(ByteList buf) throws IOException, BadDescriptorException {
         checkOpen();
         
-        return internalWrite(ByteBuffer.wrap(buf.unsafeBytes(), buf.begin(), buf.length()));
+        return internalWrite(ByteBuffer.wrap(buf.getUnsafeBytes(), buf.begin(), buf.length()));
     }
 
     /**
@@ -542,7 +546,7 @@ public class ChannelDescriptor {
     public int write(ByteList buf, int offset, int len) throws IOException, BadDescriptorException {
         checkOpen();
         
-        return internalWrite(ByteBuffer.wrap(buf.unsafeBytes(), buf.begin()+offset, len));
+        return internalWrite(ByteBuffer.wrap(buf.getUnsafeBytes(), buf.begin()+offset, len));
     }
     
     /**
@@ -624,13 +628,13 @@ public class ChannelDescriptor {
                 }
 
                 JarFile jf = new JarFile(filePath);
-                ZipEntry zf = jf.getEntry(internalPath);
+                ZipEntry entry = RubyFile.getFileEntry(jf, internalPath);
 
-                if(zf == null) {
+                if (entry == null) {
                     throw new FileNotFoundException(path);
                 }
 
-                InputStream is = jf.getInputStream(zf);
+                InputStream is = jf.getInputStream(entry);
                 // FIXME: don't use RubyIO for this
                 return new ChannelDescriptor(Channels.newChannel(is), RubyIO.getNewFileno(), flags, new FileDescriptor());
             } else {
@@ -640,6 +644,11 @@ public class ChannelDescriptor {
                 // FIXME: don't use RubyIO for this
                 return new ChannelDescriptor(Channels.newChannel(is), RubyIO.getNewFileno(), flags, new FileDescriptor());
             }
+        } else if (path.startsWith("classpath:/")) {
+            path = path.substring("classpath:/".length());
+            InputStream is = ByteList.EMPTY_BYTELIST.getClass().getClassLoader().getResourceAsStream(path);
+            // FIXME: don't use RubyIO for this
+            return new ChannelDescriptor(Channels.newChannel(is), RubyIO.getNewFileno(), flags, new FileDescriptor());
         } else {
             JRubyFile theFile = JRubyFile.create(cwd,path);
 
@@ -651,7 +660,23 @@ public class ChannelDescriptor {
                 if (theFile.exists() && flags.isExclusive()) {
                     throw new FileExistsException(path);
                 }
-                fileCreated = theFile.createNewFile();
+                try {
+                    fileCreated = theFile.createNewFile();
+                } catch (IOException ioe) {
+                    // See JRUBY-4380.
+                    // MRI behavior: raise Errno::ENOENT in case
+                    // when the directory for the file doesn't exist.
+                    // Java in such cases just throws IOException.
+                    File parent = theFile.getParentFile();
+                    if (parent != null && parent != theFile && !parent.exists()) {
+                        throw new FileNotFoundException(path);
+                    } else if (!theFile.canWrite()) {
+                        throw new PermissionDeniedException(path);
+                    } else {
+                        // for all other IO errors, just re-throw the original exception
+                        throw ioe;
+                    }
+                }
             } else {
                 if (!theFile.exists()) {
                     throw new FileNotFoundException(path);
