@@ -1,14 +1,22 @@
-######################################################################
+###############################################################################
 # tc_concat.rb
 #
-# Test suite for the Array#concat instance method.
-######################################################################
+# Test suite for the Array#concat instance method. I've added a class with
+# a custom to_ary method to ensure that it behaves properly with Array#concat.
+###############################################################################
 require "test/unit"
 
 class TC_Array_Concat_Instance < Test::Unit::TestCase
+   class AConcat
+      def to_ary
+         ['a', 'b', 'c']
+      end
+   end
+
    def setup
       @array1 = [1,2,3]
       @array2 = [4,5,6]
+      @custom = AConcat.new
    end
 
    def test_concat_basic
@@ -17,7 +25,8 @@ class TC_Array_Concat_Instance < Test::Unit::TestCase
    end
 
    def test_concat_results
-      assert_equal([1,2,3,4,5,6], @array1.concat(@array2))
+      assert_equal([1, 2, 3, 4, 5, 6], @array1.concat(@array2))
+      assert_equal([1, 2, 3, 4, 5, 6, 'a', 'b', 'c'], @array1.concat(@custom))
    end
 
    def test_concat_slice
@@ -40,5 +49,6 @@ class TC_Array_Concat_Instance < Test::Unit::TestCase
    def teardown
       @array1 = nil
       @array2 = nil
+      @custom = nil
    end
 end
