@@ -534,11 +534,7 @@ class TestFile < Test::Unit::TestCase
     File.utime(time, time, filename)
     # File mtime resolution may not be sub-second on all platforms (e.g., windows)
     # allow for some slop
-    
-    # FIXME, doesn't work on Windows: JRUBY-4047
-    unless WINDOWS
-      assert((time.to_i - File.atime(filename).to_i).abs < 5)
-    end
+    assert((time.to_i - File.atime(filename).to_i).abs < 5)
     assert((time.to_i - File.mtime(filename).to_i).abs < 5)
     File.unlink(filename)
   end
@@ -549,10 +545,7 @@ class TestFile < Test::Unit::TestCase
     time = File.mtime(filename)
     sleep 2
     File.utime(nil, nil, filename)
-    # FIXME, doesn't work on Windows: JRUBY-4047
-    unless WINDOWS
-      assert((File.atime(filename).to_i - time.to_i) >= 2)
-    end
+    assert((File.atime(filename).to_i - time.to_i) >= 2)
     assert((File.mtime(filename).to_i - time.to_i) >= 2)
     File.unlink(filename)
   end
@@ -927,6 +920,29 @@ class TestFile < Test::Unit::TestCase
     end
   end
 
+  # JRUBY-4537: File.open raises Errno::ENOENT instead of Errno::EACCES
+  def test_write_open_permission_denied
+    t = Tempfile.new('tmp' + File.basename(__FILE__))
+    t.close
+    File.open(t.path, 'w') {}
+    File.chmod(0555, t.path)
+    # jruby 1.4 raises ENOENT here
+    assert_raises(Errno::EACCES) do
+      File.open(t.path, 'w') {}
+    end
+    # jruby 1.4 raises ENOENT here
+    assert_raises(Errno::EACCES) do
+      File.open(t.path, File::WRONLY) {}
+    end
+  end
+
+  #JRUBY-4380: File.open raises IOError instead of Errno::ENOENT
+  def test_open_with_nonexisting_directory
+    file_path = "/foo/bar"
+    assert(!File.exist?(file_path))
+    assert_raises(Errno::ENOENT) { File.open(file_path, "wb") }
+  end
+
   # JRUBY-3634: File.read or File.open with a url to a file resource fails with StringIndexOutOfBounds exception
   def test_file_url
     path = File.expand_path(__FILE__)
@@ -940,4 +956,26 @@ class TestFile < Test::Unit::TestCase
     filename = 'dir/a ⼈〉〃〄⨶↖.pdf'
     assert_equal("a \342\274\210\343\200\211\343\200\203\343\200\204\342\250\266\342\206\226.pdf", File.basename(filename))
   end
+
+  #JRUBY-4387, JRUBY-4416
+  def test_file_gets_separator
+    filename = 'gets.out'
+    begin
+      File.open(filename, "wb") do |file|
+        file.print "this is a test\xFFit is only a test\ndoes it work?"
+      end
+
+      file = File.open("gets.out", "rb") do |file|
+        assert_equal("this is a test\377", file.gets("\xFF"))
+      end
+    ensure
+      File.unlink(filename)
+    end
+  end
+
+  def test_file_stat_with_missing_path
+    assert_raise(Errno::ENOENT) {
+      File::Stat.new("file:" + File.expand_path("test/test_jar2.jar") + "!/foo_bar.rb").file?
+    }
+  end
 end
